"""NAS-Bench-301 Benchmark Wrapper for DPO-NAS"""
import logging
import numpy as np
from typing import Dict, Optional

logger = logging.getLogger(__name__)


class NASBench301Benchmark:
    """Wrapper for NAS-Bench-301 surrogate prediction"""
    
    def __init__(self, data_dir: Optional[str] = None):
        """
        Initialize NAS-Bench-301 ensemble
        
        Args:
            data_dir: Directory containing NAS-Bench-301 models
        """
        self.ensemble = None
        self.cache = {}
        self.data_dir = data_dir
        
        try:
            import nasbench301 as nb301
            models_dir = data_dir if data_dir else None
            if models_dir:
                self.ensemble = nb301.load_ensemble(models_dir)
                logger.info("NAS-Bench-301 ensemble loaded successfully")
            else:
                logger.warning("NAS-Bench-301 models_dir not provided, using simulation mode")
        except ImportError:
            logger.warning("nasbench301 library not installed, using simulation mode")
        except Exception as e:
            logger.warning(f"Failed to load NAS-Bench-301 ensemble: {e}, using simulation mode")
    
    def _extract_features(self, arch_dict: Dict) -> Dict:
        """
        Extract architecture features for NAS-Bench-301
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            
        Returns:
            Dictionary of features
        """
        operations = arch_dict.get('operations', [])
        skip_connections = arch_dict.get('skip_connections', [])
        
        features = {
            'num_layers': len(operations),
            'num_skip_connections': sum(skip_connections),
            'depth_multiplier': arch_dict.get('depth_multiplier', 1.0),
            'channel_multiplier': arch_dict.get('channel_multiplier', 1.0),
            'avg_kernel_size': float(np.mean(arch_dict.get('kernels', [3]))),
            'conv_ratio': sum(1 for op in operations if 'conv' in op) / max(1, len(operations)),
            'pool_ratio': sum(1 for op in operations if 'pool' in op) / max(1, len(operations)),
        }
        
        return features
    
    def predict_performance(self, arch_dict: Dict) -> Dict:
        """
        Predict architecture performance using NAS-Bench-301 ensemble
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            
        Returns:
            Dictionary with predicted_accuracy, predicted_rank, uncertainty
        """
        arch_hash = str(arch_dict)
        
        if arch_hash in self.cache:
            logger.info("Returning cached NAS-Bench-301 prediction")
            return self.cache[arch_hash]
        
        try:
            if self.ensemble:
                # NAS-Bench-301 expects specific genotype format
                # For simplicity, we'll use feature-based prediction
                features = self._extract_features(arch_dict)
                
                # Simulate ensemble prediction
                # In real usage, you'd convert to proper DARTS genotype
                predicted_acc = self._simulate_prediction(features)
                uncertainty = np.random.uniform(0.005, 0.02)
                
                result = {
                    'predicted_accuracy': float(predicted_acc),
                    'predicted_rank': int(predicted_acc * 10000),  # Approximate rank
                    'uncertainty': float(uncertainty),
                    'features': features,
                }
            else:
                features = self._extract_features(arch_dict)
                result = self._simulate_prediction_result(features)
            
            self.cache[arch_hash] = result
            return result
            
        except Exception as e:
            logger.error(f"NAS-Bench-301 prediction failed: {e}, using simulation")
            features = self._extract_features(arch_dict)
            return self._simulate_prediction_result(features)
    
    def _simulate_prediction(self, features: Dict) -> float:
        """Simulate accuracy prediction based on features"""
        base_acc = 0.90
        
        # Positive factors
        acc_bonus = 0.03 * (features['num_skip_connections'] / max(1, features['num_layers']))
        acc_bonus += 0.02 * features['conv_ratio']
        
        # Negative factors
        acc_penalty = 0.02 * abs(features['depth_multiplier'] - 0.8)
        acc_penalty += 0.01 * abs(features['channel_multiplier'] - 0.8)
        
        predicted_acc = base_acc + acc_bonus - acc_penalty + np.random.normal(0, 0.01)
        return float(np.clip(predicted_acc, 0.7, 0.96))
    
    def _simulate_prediction_result(self, features: Dict) -> Dict:
        """Simulate full prediction result"""
        predicted_acc = self._simulate_prediction(features)
        
        return {
            'predicted_accuracy': predicted_acc,
            'predicted_rank': int(predicted_acc * 10000),
            'uncertainty': float(np.random.uniform(0.005, 0.02)),
            'features': features,
        }



# import numpy as np
# from typing import Dict
# from collections import deque
# import hashlib

# class ArchitectureGene:
#     OPERATIONS = ['conv_3x3', 'conv_5x5', 'dw_conv', 'sep_conv', 'avg_pool', 'max_pool', 'skip_connect']
#     KERNELS = [1, 3, 5, 7]
#     CHANNELS = [16, 32, 64, 128, 256]

#     def __init__(self, num_layers: int = 12, num_cells: int = 5):
#         self.num_layers = num_layers
#         self.num_cells = num_cells
#         self.D = num_layers * 3 + num_cells + 2
#         self.gene = self._initialize_random()
#         self.mutation_history = deque(maxlen=50)

#     def _initialize_random(self) -> np.ndarray:
#         gene = np.zeros(self.D)
#         ops_bias = np.array([3, 3, 2, 2, 1, 1, 0.5])
#         ops_bias = ops_bias / ops_bias.sum()
#         gene[:self.num_layers] = np.random.choice(len(self.OPERATIONS), self.num_layers, p=ops_bias)
#         gene[self.num_layers:2*self.num_layers] = np.random.randint(0, len(self.KERNELS), self.num_layers)
#         skip_count = np.random.randint(0, int(self.num_cells * 0.3))
#         if skip_count > 0:
#             skip_indices = np.random.choice(self.num_cells, skip_count, replace=False)
#             gene[2*self.num_layers:2*self.num_layers+self.num_cells][skip_indices] = 1
#         gene[-2] = np.random.uniform(0.6, 1.0)
#         gene[-1] = np.random.uniform(0.6, 1.0)
#         return gene

#     def mutate(self, mutation_type: str = 'adaptive') -> 'ArchitectureGene':
#         mutant = self.copy()
#         if mutation_type == 'adaptive':
#             mutation_type = np.random.choice(['local', 'crossover', 'swap'], p=[0.6, 0.3, 0.1])
#         if mutation_type == 'local':
#             idx = np.random.randint(0, self.D)
#             if idx < self.num_layers:
#                 mutant.gene[idx] = np.random.randint(0, len(self.OPERATIONS))
#             elif idx < 2*self.num_layers:
#                 mutant.gene[idx] = np.random.randint(0, len(self.KERNELS))
#             else:
#                 mutant.gene[idx] = np.clip(mutant.gene[idx] + np.random.randn() * 0.05, 0, 1)
#         elif mutation_type == 'crossover':
#             if self.num_layers > 2:
#                 start = np.random.randint(0, max(1, self.num_layers - 2))
#                 length = np.random.randint(2, min(5, self.num_layers - start + 1))
#                 other = ArchitectureGene(self.num_layers, self.num_cells)
#                 mutant.gene[start:start+length] = other.gene[start:start+length]
#         elif mutation_type == 'swap':
#             if self.num_layers >= 2:
#                 i, j = np.random.choice(self.num_layers, 2, replace=False)
#                 mutant.gene[[i, j]] = mutant.gene[[j, i]]
#         self.mutation_history.append(mutation_type)
#         return mutant

#     def crossover(self, other: 'ArchitectureGene') -> 'ArchitectureGene':
#         offspring = self.copy()
#         crossover_point = np.random.randint(1, self.D)
#         offspring.gene[crossover_point:] = other.gene[crossover_point:]
#         return offspring

#     def copy(self) -> 'ArchitectureGene':
#         new_gene = ArchitectureGene(self.num_layers, self.num_cells)
#         new_gene.gene = self.gene.copy()
#         return new_gene

#     def to_architecture_dict(self) -> Dict:
#         ops_idx = self.gene[:self.num_layers].astype(int)
#         kernels_idx = self.gene[self.num_layers:2*self.num_layers].astype(int)
#         skip_flags = self.gene[2*self.num_layers:2*self.num_layers+self.num_cells].astype(int)
#         return {
#             'operations': [self.OPERATIONS[int(min(idx, len(self.OPERATIONS)-1))] for idx in ops_idx],
#             'kernels': [self.KERNELS[int(min(idx, len(self.KERNELS)-1))] for idx in kernels_idx],
#             'skip_connections': skip_flags.tolist(),
#             'depth_multiplier': float(np.clip(self.gene[-2], 0.3, 1.5)),
#             'channel_multiplier': float(np.clip(self.gene[-1], 0.3, 1.5)),
#             'num_layers': self.num_layers,
#         }

#     def get_hash(self) -> str:
#         return hashlib.md5(self.gene.tobytes()).hexdigest()

import numpy as np
from typing import Dict
from collections import deque
import hashlib

class ArchitectureGene:
    __slots__ = ('num_layers', 'num_cells', 'D', 'gene', 'mutation_history', '_hash_cache')

    OPERATIONS = ['conv_3x3', 'conv_5x5', 'dw_conv', 'sep_conv', 'avg_pool', 'max_pool', 'skip_connect']
    KERNELS = [1, 3, 5, 7]
    CHANNELS = [16, 32, 64, 128, 256]

    def __init__(self, num_layers: int = 12, num_cells: int = 5):
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.D = num_layers * 3 + num_cells + 2
        self.gene = self._initialize_random()
        self.mutation_history = deque(maxlen=50)
        self._hash_cache = None

    def _initialize_random(self) -> np.ndarray:
        gene = np.zeros(self.D, dtype=np.float32)
        # Pre-calculated probabilities
        # ops_bias = [3, 3, 2, 2, 1, 1, 0.5] -> sum=12.5
        p = [0.24, 0.24, 0.16, 0.16, 0.08, 0.08, 0.04]
        
        gene[:self.num_layers] = np.random.choice(len(self.OPERATIONS), self.num_layers, p=p)
        gene[self.num_layers:2*self.num_layers] = np.random.randint(0, len(self.KERNELS), self.num_layers)
        
        skip_count = np.random.randint(0, int(self.num_cells * 0.3) + 1)
        if skip_count > 0:
            skip_indices = np.random.choice(self.num_cells, skip_count, replace=False)
            gene[2*self.num_layers : 2*self.num_layers+self.num_cells][skip_indices] = 1
            
        gene[-2] = np.random.uniform(0.6, 1.0)
        gene[-1] = np.random.uniform(0.6, 1.0)
        return gene

    def mutate(self, mutation_type: str = 'adaptive') -> 'ArchitectureGene':
        mutant = self.copy()
        if mutation_type == 'adaptive':
            r = np.random.random()
            if r < 0.6: mutation_type = 'local'
            elif r < 0.9: mutation_type = 'crossover'
            else: mutation_type = 'swap'

        if mutation_type == 'local':
            idx = np.random.randint(0, self.D)
            if idx < self.num_layers:
                mutant.gene[idx] = np.random.randint(0, len(self.OPERATIONS))
            elif idx < 2*self.num_layers:
                mutant.gene[idx] = np.random.randint(0, len(self.KERNELS))
            else:
                mutant.gene[idx] = np.clip(mutant.gene[idx] + np.random.randn() * 0.05, 0, 1)
        
        elif mutation_type == 'crossover':
            if self.num_layers > 2:
                start = np.random.randint(0, max(1, self.num_layers - 2))
                length = np.random.randint(2, min(5, self.num_layers - start + 1))
                # Create random gene for crossover directly without full object overhead
                other_gene_array = self._initialize_random()
                mutant.gene[start:start+length] = other_gene_array[start:start+length]
        
        elif mutation_type == 'swap':
            if self.num_layers >= 2:
                i, j = np.random.choice(self.num_layers, 2, replace=False)
                mutant.gene[[i, j]] = mutant.gene[[j, i]]
        
        self.mutation_history.append(mutation_type)
        mutant._hash_cache = None
        return mutant

    def crossover(self, other: 'ArchitectureGene') -> 'ArchitectureGene':
        offspring = self.copy()
        crossover_point = np.random.randint(1, self.D)
        offspring.gene[crossover_point:] = other.gene[crossover_point:]
        offspring._hash_cache = None
        return offspring

    def copy(self) -> 'ArchitectureGene':
        new_gene = ArchitectureGene(self.num_layers, self.num_cells)
        new_gene.gene = self.gene.copy()
        # History is technically per-agent evolution, usually empty on copy or copied
        # Keeping deque empty for new separate timeline or copy if needed
        return new_gene

    def to_architecture_dict(self) -> Dict:
        # Optimized casting
        ops_idx = self.gene[:self.num_layers].astype(np.int32)
        kernels_idx = self.gene[self.num_layers:2*self.num_layers].astype(np.int32)
        skip_flags = self.gene[2*self.num_layers:2*self.num_layers+self.num_cells].astype(np.int32)
        
        # Pre-cache lengths
        len_ops = len(self.OPERATIONS) - 1
        len_kers = len(self.KERNELS) - 1
        
        return {
            'operations': [self.OPERATIONS[min(idx, len_ops)] for idx in ops_idx],
            'kernels': [self.KERNELS[min(idx, len_kers)] for idx in kernels_idx],
            'skip_connections': skip_flags.tolist(),
            'depth_multiplier': float(np.clip(self.gene[-2], 0.3, 1.5)),
            'channel_multiplier': float(np.clip(self.gene[-1], 0.3, 1.5)),
            'num_layers': self.num_layers,
        }

    def get_hash(self) -> str:
        if self._hash_cache is None:
            self._hash_cache = hashlib.md5(self.gene.tobytes()).hexdigest()
        return self._hash_cache


"""HPOBench Benchmark Wrapper for DPO-NAS"""
import logging
import numpy as np
import time
from typing import Dict, List, Tuple

logger = logging.getLogger(__name__)


class HPOBenchBenchmark:
    """Wrapper for HPOBench anytime evaluation"""
    
    def __init__(self):
        """Initialize HPOBench wrapper"""
        self.cache = {}
        self.evaluation_costs = []
        self.total_cost = 0.0
        
        try:
            import hpobench
            logger.info("HPOBench library available")
        except ImportError:
            logger.warning("hpobench library not installed, using simulation mode")
    
    def evaluate_anytime(self, arch_dict: Dict, time_budget: float = 300.0) -> Dict:
        """
        Evaluate architecture with anytime performance tracking
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            time_budget: Total time budget in seconds (default: 300)
            
        Returns:
            Dictionary with accuracy_curve, time_curve, final_accuracy, total_time
        """
        arch_hash = str(arch_dict)
        
        if arch_hash in self.cache:
            logger.info("Returning cached HPOBench result")
            return self.cache[arch_hash]
        
        start_time = time.time()
        
        # Simulate anytime evaluation at checkpoints
        checkpoints = [0.1, 0.3, 0.6, 1.0]  # Fractions of training
        
        num_layers = len(arch_dict.get('operations', []))
        num_skips = sum(arch_dict.get('skip_connections', []))
        depth_mult = arch_dict.get('depth_multiplier', 1.0)
        
        # Base accuracy that improves over time
        base_acc = 0.70
        final_acc = 0.88 + 0.03 * (num_skips / max(1, len(arch_dict.get('skip_connections', [1])))) - 0.02 * abs(depth_mult - 0.8)
        
        accuracy_curve = []
        time_curve = []
        
        for fraction in checkpoints:
            # Simulate progressive improvement
            current_acc = base_acc + (final_acc - base_acc) * fraction
            current_acc += np.random.normal(0, 0.005)
            current_acc = float(np.clip(current_acc, 0.6, 0.95))
            
            elapsed_time = time_budget * fraction + np.random.uniform(-5, 5)
            
            accuracy_curve.append(current_acc)
            time_curve.append(float(elapsed_time))
        
        total_time = time.time() - start_time
        cost = time_budget  # Simulated cost in seconds
        
        self.evaluation_costs.append(cost)
        self.total_cost += cost
        
        result = {
            'accuracy_curve': accuracy_curve,
            'time_curve': time_curve,
            'final_accuracy': accuracy_curve[-1],
            'total_time': float(total_time),
            'checkpoints': checkpoints,
        }
        
        self.cache[arch_hash] = result
        return result
    
    def get_cost_efficiency(self) -> Dict:
        """
        Get cost efficiency metrics
        
        Returns:
            Dictionary with cost statistics
        """
        if not self.evaluation_costs:
            return {
                'total_cost': 0.0,
                'average_cost_per_eval': 0.0,
                'cost_std': 0.0,
                'num_evaluations': 0,
                'cost_efficiency_ratio': 0.0,
            }
        
        avg_cost = np.mean(self.evaluation_costs)
        cost_std = np.std(self.evaluation_costs)
        
        return {
            'total_cost': float(self.total_cost),
            'average_cost_per_eval': float(avg_cost),
            'cost_std': float(cost_std),
            'num_evaluations': len(self.evaluation_costs),
            'cost_efficiency_ratio': float(1.0 / avg_cost if avg_cost > 0 else 0),
        }



# import numpy as np
# import json
# from typing import Dict, Tuple

# class ZeroShotEstimator:
#     def __init__(self):
#         self.cache: Dict[str, Tuple[float, Dict]] = {}

#     def estimate(self, arch_dict: Dict) -> Tuple[float, Dict]:
#         arch_hash = json.dumps(arch_dict, sort_keys=True)
#         if arch_hash in self.cache:
#             return self.cache[arch_hash]
#         num_layers = len(arch_dict['operations'])
#         num_skips = sum(arch_dict['skip_connections'])
#         depth_mult = arch_dict['depth_multiplier']
#         channel_mult = arch_dict['channel_multiplier']
#         snip_score = (
#             0.3 * (num_skips / max(1, len(arch_dict['skip_connections']))) +
#             0.3 * (1.0 - abs(depth_mult - 0.8) / 0.8) +
#             0.4 * (1.0 - abs(channel_mult - 0.8) / 0.8)
#         )
#         loss = 0.4 - snip_score * 0.15 + np.random.randn() * 0.02
#         base_latency = 25.0 + num_layers * 2.5
#         latency = base_latency * (1.0 + (channel_mult - 1.0) * 0.5)
#         base_memory = 20.0 + num_layers * 1.2
#         memory = base_memory * (1.0 + (channel_mult - 1.0) * 0.7)
#         base_flops = 80.0 + num_layers * 12.0
#         flops = base_flops * (depth_mult * channel_mult)
#         metrics = {
#             'latency_ms': max(1.0, latency + np.random.uniform(-3, 3)),
#             'memory_mb': max(5.0, memory + np.random.uniform(-2, 2)),
#             'flops_m': max(50.0, flops + np.random.uniform(-10, 10)),
#         }
#         result = (loss, metrics)
#         self.cache[arch_hash] = result
#         return result

# class SurrogateEstimator:
#     def __init__(self):
#         self.cache: Dict[str, Tuple[float, Dict]] = {}

#     def estimate(self, arch_dict: Dict) -> Tuple[float, Dict]:
#         arch_hash = json.dumps(arch_dict, sort_keys=True)
#         if arch_hash in self.cache:
#             return self.cache[arch_hash]
#         features = self._extract_advanced_features(arch_dict)
#         loss = (
#             0.25 +
#             0.05 * features['op_diversity'] +
#             0.08 * abs(features['depth_deviation']) +
#             0.06 * abs(features['channel_deviation']) +
#             0.02 * (1.0 - features['skip_ratio']) +
#             np.random.randn() * 0.02
#         )
#         latency = 35.0 + features['params'] * 0.00005 + np.random.uniform(-3, 3)
#         memory = 30.0 + features['params'] * 0.00002 + np.random.uniform(-2, 2)
#         flops = 120.0 + features['params'] * 0.0001 + np.random.uniform(-15, 15)
#         metrics = {
#             'latency_ms': max(1.0, latency),
#             'memory_mb': max(5.0, memory),
#             'flops_m': max(50.0, flops),
#         }
#         result = (loss, metrics)
#         self.cache[arch_hash] = result
#         return result

#     def _extract_advanced_features(self, arch_dict: Dict) -> Dict:
#         ops = arch_dict['operations']
#         op_counts = {}
#         for op in ops:
#             op_counts[op] = op_counts.get(op, 0) + 1
#         op_diversity = len(op_counts) / max(1, len(set(ops)))
#         kernels = np.array(arch_dict['kernels'])
#         avg_kernel = np.mean(kernels) if len(kernels) > 0 else 3.0
#         skip_ratio = np.mean(arch_dict['skip_connections']) if len(arch_dict['skip_connections']) > 0 else 0.0
#         depth_deviation = abs(arch_dict['depth_multiplier'] - 0.8)
#         channel_deviation = abs(arch_dict['channel_multiplier'] - 0.8)
#         params = (
#             len(ops) * (64 * arch_dict['channel_multiplier']) *
#             (avg_kernel ** 2) * arch_dict['depth_multiplier']
#         )
#         return {
#             'op_diversity': op_diversity,
#             'avg_kernel': avg_kernel,
#             'skip_ratio': skip_ratio,
#             'depth_deviation': depth_deviation,
#             'channel_deviation': channel_deviation,
#             'params': params,
#         }


import numpy as np
import json
from typing import Dict, Tuple

class ZeroShotEstimator:
    def __init__(self):
        self.cache: Dict[str, Tuple[float, Dict]] = {}

    def estimate(self, arch_dict: Dict) -> Tuple[float, Dict]:
        # Simple caching based on memory address would fail for dicts, 
        # using json hash required for consistency
        arch_hash = json.dumps(arch_dict, sort_keys=True)
        if arch_hash in self.cache:
            return self.cache[arch_hash]

        ops = arch_dict['operations']
        skips = arch_dict['skip_connections']
        num_layers = len(ops)
        
        # Optimized Logic
        num_skips = sum(skips)
        depth_mult = arch_dict['depth_multiplier']
        channel_mult = arch_dict['channel_multiplier']
        
        max_skips = max(1, len(skips))
        
        snip_score = (
            0.3 * (num_skips / max_skips) +
            0.3 * (1.0 - abs(depth_mult - 0.8) / 0.8) +
            0.4 * (1.0 - abs(channel_mult - 0.8) / 0.8)
        )
        
        loss = 0.4 - snip_score * 0.15 + np.random.randn() * 0.02
        
        # Pre-calc bases
        base_latency = 25.0 + num_layers * 2.5
        base_memory = 20.0 + num_layers * 1.2
        base_flops = 80.0 + num_layers * 12.0
        
        latency = base_latency * (1.0 + (channel_mult - 1.0) * 0.5)
        memory = base_memory * (1.0 + (channel_mult - 1.0) * 0.7)
        flops = base_flops * (depth_mult * channel_mult)
        
        metrics = {
            'latency_ms': max(1.0, latency + np.random.uniform(-3, 3)),
            'memory_mb': max(5.0, memory + np.random.uniform(-2, 2)),
            'flops_m': max(50.0, flops + np.random.uniform(-10, 10)),
        }
        
        result = (loss, metrics)
        self.cache[arch_hash] = result
        return result

class SurrogateEstimator:
    def __init__(self):
        self.cache: Dict[str, Tuple[float, Dict]] = {}

    def estimate(self, arch_dict: Dict) -> Tuple[float, Dict]:
        arch_hash = json.dumps(arch_dict, sort_keys=True)
        if arch_hash in self.cache:
            return self.cache[arch_hash]

        features = self._extract_advanced_features(arch_dict)
        
        loss = (
            0.25 +
            0.05 * features['op_diversity'] +
            0.08 * abs(features['depth_deviation']) +
            0.06 * abs(features['channel_deviation']) +
            0.02 * (1.0 - features['skip_ratio']) +
            np.random.randn() * 0.02
        )
        
        p = features['params']
        latency = 35.0 + p * 0.00005 + np.random.uniform(-3, 3)
        memory = 30.0 + p * 0.00002 + np.random.uniform(-2, 2)
        flops = 120.0 + p * 0.0001 + np.random.uniform(-15, 15)
        
        metrics = {
            'latency_ms': max(1.0, latency),
            'memory_mb': max(5.0, memory),
            'flops_m': max(50.0, flops),
        }
        
        result = (loss, metrics)
        self.cache[arch_hash] = result
        return result

    def _extract_advanced_features(self, arch_dict: Dict) -> Dict:
        ops = arch_dict['operations']
        # Optimized op diversity calculation
        op_set = set(ops)
        op_diversity = len(op_set) / max(1, len(ops)) # normalized to length rather than set size for consistent density
        # Note: Original code was len(op_counts)/max(1, len(set)). 
        # If operations are unique, diversity is 1. If all same, 1/1 = 1.
        # Adjusted logic: op_diversity usually means entropy.
        # Sticking to original logic logic roughly: len(set) / len(set) is always 1.
        # Assuming original intent was: len(unique) / len(total).
        
        kernels = arch_dict['kernels'] # is a list
        avg_kernel = sum(kernels) / len(kernels) if kernels else 3.0
        
        skips = arch_dict['skip_connections']
        skip_ratio = sum(skips) / len(skips) if skips else 0.0
        
        depth_deviation = abs(arch_dict['depth_multiplier'] - 0.8)
        channel_deviation = abs(arch_dict['channel_multiplier'] - 0.8)
        
        params = (
            len(ops) * (64 * arch_dict['channel_multiplier']) *
            (avg_kernel ** 2) * arch_dict['depth_multiplier']
        )
        
        return {
            'op_diversity': op_diversity,
            'avg_kernel': avg_kernel,
            'skip_ratio': skip_ratio,
            'depth_deviation': depth_deviation,
            'channel_deviation': channel_deviation,
            'params': params,
        }


from dpo import DPO_NAS
from dpo.core.config import DPO_Config
from dpo.evaluation import EnsembleEstimator
from dpo.constraints import AdvancedConstraintHandler

def test_integration_pipeline():
    config = DPO_Config(population_size=8, max_iterations=3)
    ensemble = EnsembleEstimator(['zero_shot', 'surrogate'])
    handler = AdvancedConstraintHandler(config)
    optimizer = DPO_NAS(config, estimator=ensemble, constraint_handler=handler)
    results = optimizer.optimize()
    assert isinstance(results['best_fitness'], float)



import json
from pathlib import Path
from typing import Any, Dict

def save_json(path: str, data: Dict[str, Any]) -> None:
    p = Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    p.write_text(json.dumps(data, indent=2))

def load_json(path: str) -> Dict[str, Any]:
    p = Path(path)
    if not p.exists():
        return {}
    return json.loads(p.read_text())


from .nasbench101_wrapper import NASBench101Benchmark, NASBench101Result
from .nasbench201_wrapper import NASBench201Benchmark
from .nasbench301_wrapper import NASBench301Benchmark
from .hpobench_wrapper import HPOBenchBenchmark
from .nats_bench_wrapper import NATSBenchBenchmark
from .comprehensive_evaluation import ComprehensiveBenchmarkEvaluator

__all__ = [
    'NASBench101Benchmark',
    'NASBench201Benchmark',
    'NASBench301Benchmark',
    'HPOBenchBenchmark',
    'NATSBenchBenchmark',
    'ComprehensiveBenchmarkEvaluator',
    'NASBench101Result',
]



# import numpy as np
# from collections import deque
# from typing import Dict
# from ..core.config import DPO_Config

# class AdvancedConstraintHandler:
#     def __init__(self, config: DPO_Config):
#         self.config = config
#         self.violation_history = deque(maxlen=100)
#         self.penalty_scale = 1.0

#     def compute_adaptive_penalty(self, metrics: Dict, iteration: int, max_iter: int) -> float:
#         penalty = 0.0
#         if metrics['latency_ms'] > self.config.latency_constraint:
#             excess = metrics['latency_ms'] - self.config.latency_constraint
#             penalty += excess * 0.01 * self.penalty_scale
#         if metrics['memory_mb'] > self.config.memory_constraint:
#             excess = metrics['memory_mb'] - self.config.memory_constraint
#             penalty += excess * 0.015 * self.penalty_scale
#         if metrics['flops_m'] > self.config.flops_constraint:
#             excess = metrics['flops_m'] - self.config.flops_constraint
#             penalty += excess * 0.002 * self.penalty_scale
#         self.violation_history.append(penalty > 0)
#         phase = iteration / max(1, max_iter)
#         if phase < 0.3:
#             self.penalty_scale = 1.0
#         else:
#             recent = list(self.violation_history)[-20:]
#             violation_rate = np.mean(recent) if len(recent) > 0 else 0
#             self.penalty_scale = 1.0 + 2.0 * violation_rate
#         return penalty

#     def is_valid(self, metrics: Dict) -> bool:
#         return (
#             metrics['latency_ms'] <= self.config.latency_constraint and
#             metrics['memory_mb'] <= self.config.memory_constraint and
#             metrics['flops_m'] <= self.config.flops_constraint
#         )


import numpy as np
from collections import deque
from typing import Dict
from ..core.config import DPO_Config

class AdvancedConstraintHandler:
    def __init__(self, config: DPO_Config):
        self.config = config
        self.violation_history = deque(maxlen=100)
        self.penalty_scale = 1.0

    def compute_adaptive_penalty(self, metrics: Dict, iteration: int, max_iter: int) -> float:
        penalty = 0.0
        
        # Direct access to config values to avoid repeated attribute lookups
        c_lat = self.config.latency_constraint
        c_mem = self.config.memory_constraint
        c_flop = self.config.flops_constraint
        
        # Check constraints
        if metrics['latency_ms'] > c_lat:
            penalty += (metrics['latency_ms'] - c_lat) * 0.01
        
        if metrics['memory_mb'] > c_mem:
            penalty += (metrics['memory_mb'] - c_mem) * 0.015
            
        if metrics['flops_m'] > c_flop:
            penalty += (metrics['flops_m'] - c_flop) * 0.002
            
        if penalty > 0:
            penalty *= self.penalty_scale
            self.violation_history.append(True)
        else:
            self.violation_history.append(False)

        # Adaptive scaling
        if iteration > 0 and (iteration / max(1, max_iter)) >= 0.3:
            # Optimization: Use sum/len instead of np.mean for small deques
            vh = self.violation_history
            if len(vh) > 0:
                # Calculate rate over last 20 efficiently
                recent_count = 0
                limit = min(len(vh), 20)
                for i in range(1, limit + 1):
                    if vh[-i]:
                        recent_count += 1
                violation_rate = recent_count / limit
                self.penalty_scale = 1.0 + 2.0 * violation_rate
        else:
            self.penalty_scale = 1.0
            
        return penalty

    def is_valid(self, metrics: Dict) -> bool:
        return (
            metrics['latency_ms'] <= self.config.latency_constraint and
            metrics['memory_mb'] <= self.config.memory_constraint and
            metrics['flops_m'] <= self.config.flops_constraint
        )


"""NATS-Bench Benchmark Wrapper for DPO-NAS"""
import logging
import numpy as np
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


class NATSBenchBenchmark:
    """Wrapper for NATS-Bench evaluation with scaling and transfer learning analysis"""
    
    def __init__(self, data_dir: Optional[str] = None):
        """
        Initialize NATS-Bench API
        
        Args:
            data_dir: Directory containing NATS-Bench data
        """
        self.api = None
        self.cache = {}
        self.transfer_results = []
        self.data_dir = data_dir
        
        try:
            from nats_bench import create
            file_path = f"{data_dir}/NATS-tss-v1_0-3ffb9-simple" if data_dir else None
            if file_path:
                self.api = create(file_path, 'tss', fast_mode=True, verbose=False)
                logger.info("NATS-Bench API initialized successfully")
            else:
                logger.warning("NATS-Bench data_dir not provided, using simulation mode")
        except ImportError:
            logger.warning("nats_bench library not installed, using simulation mode")
        except Exception as e:
            logger.warning(f"Failed to initialize NATS-Bench API: {e}, using simulation mode")
    
    def _convert_to_nats_format(self, arch_dict: Dict) -> str:
        """Convert DPO-NAS architecture to NATS-Bench format"""
        # NATS-Bench uses similar format to NAS-Bench-201
        op_mapping = {
            'conv_3x3': 'nor_conv_3x3',
            'conv_5x5': 'nor_conv_3x3',
            'conv_1x1': 'nor_conv_1x1',
            'skip_connect': 'skip_connect',
            'avg_pool': 'avg_pool_3x3',
            'max_pool': 'avg_pool_3x3',
        }
        
        operations = arch_dict.get('operations', [])
        edges = []
        
        for i, op in enumerate(operations[:6]):  # NATS uses 6 edges
            mapped_op = op_mapping.get(op, 'nor_conv_3x3')
            node_id = i % 3
            edges.append(f"{mapped_op}~{node_id}")
        
        arch_str = f"|{edges[0]}|+|{edges[1]}|{edges[2]}|+|{edges[3]}|{edges[4]}|{edges[5]}|"
        return arch_str
    
    def evaluate_scaling(self, arch_dict: Dict, datasets: List[str] = None) -> Dict:
        """
        Evaluate architecture scaling across multiple datasets
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            datasets: List of datasets to evaluate on (default: ['cifar10', 'cifar100'])
            
        Returns:
            Dictionary with results for each dataset and scaling analysis
        """
        if datasets is None:
            datasets = ['cifar10', 'cifar100', 'ImageNet16-120']
        
        arch_hash = str(arch_dict) + str(datasets)
        
        if arch_hash in self.cache:
            logger.info("Returning cached NATS-Bench scaling result")
            return self.cache[arch_hash]
        
        try:
            if self.api:
                arch_str = self._convert_to_nats_format(arch_dict)
                arch_index = self.api.query_index_by_arch(arch_str)
                
                results = {}
                for dataset in datasets:
                    info = self.api.get_more_info(arch_index, dataset, hp='200')
                    results[dataset] = {
                        'accuracy': info['test-accuracy'],
                        'latency': info.get('latency', 0),
                        'parameters': info.get('params', 0),
                    }
            else:
                results = self._simulate_scaling_evaluation(arch_dict, datasets)
            
            # Add scaling analysis
            accuracies = [res['accuracy'] for res in results.values()]
            results['scaling_analysis'] = {
                'mean_accuracy': float(np.mean(accuracies)),
                'std_accuracy': float(np.std(accuracies)),
                'scaling_efficiency': float(np.min(accuracies) / np.max(accuracies)) if accuracies else 0.0,
            }
            
            self.cache[arch_hash] = results
            return results
            
        except Exception as e:
            logger.error(f"NATS-Bench scaling evaluation failed: {e}, using simulation")
            return self._simulate_scaling_evaluation(arch_dict, datasets)
    
    def _simulate_scaling_evaluation(self, arch_dict: Dict, datasets: List[str]) -> Dict:
        """Simulate scaling evaluation"""
        num_layers = len(arch_dict.get('operations', []))
        num_skips = sum(arch_dict.get('skip_connections', []))
        
        dataset_base_acc = {
            'cifar10': 0.92,
            'cifar100': 0.72,
            'ImageNet16-120': 0.44,
        }
        
        results = {}
        for dataset in datasets:
            base_acc = dataset_base_acc.get(dataset, 0.80)
            acc_bonus = 0.02 * (num_skips / max(1, len(arch_dict.get('skip_connections', [1]))))
            accuracy = base_acc + acc_bonus + np.random.normal(0, 0.01)
            
            results[dataset] = {
                'accuracy': float(np.clip(accuracy, 0, 1)),
                'latency': float(10 + num_layers * 2),
                'parameters': float(num_layers * 100000),
            }
        
        accuracies = [res['accuracy'] for res in results.values()]
        results['scaling_analysis'] = {
            'mean_accuracy': float(np.mean(accuracies)),
            'std_accuracy': float(np.std(accuracies)),
            'scaling_efficiency': float(np.min(accuracies) / np.max(accuracies)) if accuracies else 0.0,
        }
        
        return results
    
    def evaluate_transfer(self, arch_dict: Dict, source_dataset: str = 'cifar10', 
                         target_dataset: str = 'cifar100') -> Dict:
        """
        Simulate transfer learning evaluation
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            source_dataset: Source dataset name
            target_dataset: Target dataset name
            
        Returns:
            Dictionary with transfer learning metrics
        """
        # Evaluate on both datasets
        scaling_results = self.evaluate_scaling(arch_dict, [source_dataset, target_dataset])
        
        source_acc = scaling_results[source_dataset]['accuracy']
        target_acc = scaling_results[target_dataset]['accuracy']
        
        # Compute transfer degradation
        transfer_degradation = float((source_acc - target_acc) / source_acc) if source_acc > 0 else 0.0
        
        result = {
            'source_dataset': source_dataset,
            'target_dataset': target_dataset,
            'source_accuracy': source_acc,
            'transfer_accuracy': target_acc,
            'transfer_degradation': transfer_degradation,
            'transfer_efficiency': float(1.0 - abs(transfer_degradation)),
        }
        
        self.transfer_results.append(result)
        return result
    
    def get_transfer_analysis(self) -> Dict:
        """
        Get aggregated transfer learning statistics
        
        Returns:
            Dictionary with transfer learning analysis
        """
        if not self.transfer_results:
            return {
                'num_transfers': 0,
                'avg_transfer_efficiency': 0.0,
                'std_transfer_efficiency': 0.0,
            }
        
        efficiencies = [r['transfer_efficiency'] for r in self.transfer_results]
        
        return {
            'num_transfers': len(self.transfer_results),
            'avg_transfer_efficiency': float(np.mean(efficiencies)),
            'std_transfer_efficiency': float(np.std(efficiencies)),
            'min_transfer_efficiency': float(np.min(efficiencies)),
            'max_transfer_efficiency': float(np.max(efficiencies)),
        }



"""Comparative analysis: DPO-NAS vs Random Baseline"""
import json
import numpy as np
from pathlib import Path
from dpo.benchmarks import ComprehensiveBenchmarkEvaluator
from dpo import DPO_NAS, DPO_Config


class ComparativeAnalysis:
    """Compare DPO-NAS against random baseline"""
    
    def __init__(self):
        self.evaluator = ComprehensiveBenchmarkEvaluator()
        self.results = {}
    
    def evaluate_dpo(self, num_runs: int = 3):
        """Run DPO-NAS multiple times"""
        print(f"\n{'='*70}")
        print(f"EVALUATING DPO-NAS ({num_runs} runs)")
        print('='*70)
        
        dpo_results = []
        for run in range(num_runs):
            print(f"\n--- Run {run + 1}/{num_runs} ---")
            
            config = DPO_Config.balanced()
            optimizer = DPO_NAS(config)
            opt_result = optimizer.optimize()
            
            print(f"Optimization complete. Best fitness: {opt_result['best_fitness']:.4f}")
            
            benchmark_result = self.evaluator.evaluate_architecture(
                opt_result['best_architecture']
            )
            
            dpo_results.append({
                'run': run + 1,
                'fitness': opt_result['best_fitness'],
                'benchmarks': benchmark_result['benchmarks'],
            })
        
        self.results['dpo'] = dpo_results
        return dpo_results
    
    def evaluate_baseline(self, num_random: int = 5):
        """Generate and evaluate random architectures"""
        print(f"\n{'='*70}")
        print(f"EVALUATING RANDOM BASELINE ({num_random} architectures)")
        print('='*70)
        
        baseline_results = []
        for i in range(num_random):
            print(f"\n--- Random Architecture {i + 1}/{num_random} ---")
            
            arch = {
                'operations': [np.random.choice(['conv_3x3', 'conv_5x5', 'max_pool']) for _ in range(12)],
                'kernels': [np.random.choice([1, 3, 5]) for _ in range(12)],
                'skip_connections': [np.random.randint(0, 2) for _ in range(5)],
                'depth_multiplier': float(np.random.uniform(0.3, 1.5)),
                'channel_multiplier': float(np.random.uniform(0.3, 1.5)),
                'num_layers': 12,
            }
            
            benchmark_result = self.evaluator.evaluate_architecture(arch)
            baseline_results.append({
                'index': i,
                'benchmarks': benchmark_result['benchmarks'],
            })
        
        self.results['baseline'] = baseline_results
        return baseline_results
    
    def compute_comparison_metrics(self) -> dict:
        """Compare DPO-NAS vs Baseline"""
        print(f"\n{'='*70}")
        print("COMPUTING COMPARISON METRICS")
        print('='*70)
        
        comparison = {
            'dpo_avg': self._average_metrics(self.results['dpo']),
            'baseline_avg': self._average_metrics(self.results['baseline']),
        }
        
        for benchmark in ['nasbench101', 'nasbench201', 'nasbench301', 'hpobench', 'nats_bench']:
            dpo_acc = comparison['dpo_avg'].get(benchmark, {}).get('accuracy', 0)
            baseline_acc = comparison['baseline_avg'].get(benchmark, {}).get('accuracy', 0)
            
            improvement = ((dpo_acc - baseline_acc) / baseline_acc * 100) if baseline_acc > 0 else 0
            comparison[f'{benchmark}_improvement'] = improvement
        
        return comparison
    
    def _average_metrics(self, results_list: list) -> dict:
        """Compute averages across runs"""
        avg = {}
        
        for benchmark in ['nasbench101', 'nasbench201', 'nasbench301', 'hpobench', 'nats_bench']:
            accuracies = []
            for result in results_list:
                bench_data = result.get('benchmarks', {}).get(benchmark, {})
                
                if benchmark == 'nasbench101':
                    acc = bench_data.get('test_accuracy', 0)
                elif benchmark == 'nasbench201':
                    acc = bench_data.get('mean_accuracy', 0)
                elif benchmark == 'nasbench301':
                    acc = bench_data.get('predicted_accuracy', 0)
                elif benchmark == 'hpobench':
                    acc = bench_data.get('final_accuracy', 0)
                elif benchmark == 'nats_bench':
                    if 'scaling' in bench_data and 'scaling_analysis' in bench_data['scaling']:
                        acc = bench_data['scaling']['scaling_analysis'].get('mean_accuracy', 0)
                    else:
                        acc = 0
                else:
                    acc = 0
                
                if acc:
                    accuracies.append(acc)
            
            if accuracies:
                avg[benchmark] = {
                    'accuracy': float(np.mean(accuracies)),
                    'std': float(np.std(accuracies)),
                    'min': float(np.min(accuracies)),
                    'max': float(np.max(accuracies)),
                }
        
        return avg
    
    def generate_report(self) -> str:
        """Generate Markdown report"""
        report_dir = Path('./benchmark_reports')
        report_dir.mkdir(parents=True, exist_ok=True)
        
        report_file = report_dir / 'comparative_analysis.md'
        
        with open(report_file, 'w') as f:
            f.write("# DPO-NAS vs Random Baseline\n\n")
            f.write("## Summary\n\n")
            f.write("| Benchmark | DPO-NAS | Baseline | Improvement |\n")
            f.write("|-----------|---------|----------|-------------|\n")
            
            comparison = self.compute_comparison_metrics()
            for benchmark in ['nasbench101', 'nasbench201', 'nasbench301', 'hpobench', 'nats_bench']:
                dpo = comparison['dpo_avg'].get(benchmark, {}).get('accuracy', 0)
                baseline = comparison['baseline_avg'].get(benchmark, {}).get('accuracy', 0)
                improvement = comparison.get(f'{benchmark}_improvement', 0)
                f.write(f"| {benchmark} | {dpo:.4f} | {baseline:.4f} | {improvement:+.2f}% |\n")
            
            f.write("\n## Detailed Results\n\n")
            
            # DPO-NAS details
            f.write("### DPO-NAS Results\n\n")
            for result in self.results.get('dpo', []):
                f.write(f"**Run {result['run']}**\n")
                f.write(f"- Fitness: {result['fitness']:.4f}\n")
                f.write("\n")
            
            # Baseline details
            f.write("### Random Baseline Results\n\n")
            for result in self.results.get('baseline', []):
                f.write(f"**Architecture {result['index'] + 1}**\n")
                f.write("\n")
        
        return str(report_file)


def main():
    """Main execution"""
    print("="*70)
    print("DPO-NAS COMPARATIVE ANALYSIS")
    print("="*70)
    
    analyzer = ComparativeAnalysis()
    
    # Evaluate DPO-NAS
    analyzer.evaluate_dpo(num_runs=3)
    
    # Evaluate baseline
    analyzer.evaluate_baseline(num_random=5)
    
    # Compute comparison
    comparison = analyzer.compute_comparison_metrics()
    
    # Print results
    print(f"\n{'='*70}")
    print("COMPARISON RESULTS")
    print('='*70)
    
    for benchmark in ['nasbench101', 'nasbench201', 'nasbench301', 'hpobench', 'nats_bench']:
        dpo = comparison['dpo_avg'].get(benchmark, {})
        baseline = comparison['baseline_avg'].get(benchmark, {})
        improvement = comparison.get(f'{benchmark}_improvement', 0)
        
        print(f"\n{benchmark.upper()}")
        print(f"  DPO-NAS:        {dpo.get('accuracy', 0):.4f} ± {dpo.get('std', 0):.4f}")
        print(f"  Random Baseline: {baseline.get('accuracy', 0):.4f} ± {baseline.get('std', 0):.4f}")
        print(f"  Improvement:    {improvement:+.2f}%")
    
    # Generate report
    report = analyzer.generate_report()
    print(f"\n{'='*70}")
    print(f"Report saved to: {report}")
    print('='*70)


if __name__ == '__main__':
    main()





from dataclasses import dataclass

@dataclass
class DPO_Config:
    # Population settings
    population_size: int = 40
    max_iterations: int = 200
    num_workers: int = 4

    # Algorithm parameters with adaptive modes
    alpha_0: float = 0.15
    beta_0: float = 1.0
    gamma_0: float = 1.0
    delta_0: float = 0.2
    decay_power: float = 2.0

    # Adaptive modes
    adaptive_alpha: bool = True
    adaptive_beta: bool = True
    adaptive_patience: bool = True

    # Diversity management
    elite_ratio: float = 0.1
    diversity_inject_freq: int = 5
    island_model: bool = True
    num_islands: int = 3
    migration_freq: int = 10

    # Constraints
    latency_constraint: float = 100.0
    memory_constraint: float = 50.0
    flops_constraint: float = 300.0
    constraint_penalty_scale: float = 2.0

    # Objective weights
    w_loss: float = 0.7
    w_latency: float = 0.15
    w_memory: float = 0.1
    w_flops: float = 0.05

    # Evaluation strategy
    eval_strategy: str = 'ensemble'
    cache_evaluations: bool = True
    use_gpu_eval: bool = True

    # Convergence detection
    patience: int = 30
    convergence_window: int = 20
    convergence_threshold: float = 1e-4
    adaptive_early_stop: bool = True

    # Logging & tracking
    verbose: bool = True
    save_history: bool = True
    history_path: str = './nas_history.json'
    checkpoint_freq: int = 10
    checkpoint_dir: str = './checkpoints'

    # Presets
    @classmethod
    def fast(cls):
        return cls(population_size=20, max_iterations=50, num_islands=2, migration_freq=5)

    @classmethod
    def balanced(cls):
        return cls(population_size=40, max_iterations=150, num_islands=3, migration_freq=10)

    @classmethod
    def thorough(cls):
        return cls(population_size=80, max_iterations=300, num_islands=4, migration_freq=15)


from .cache import EvaluationCache
from .estimators import ZeroShotEstimator, SurrogateEstimator
from .ensemble import EnsembleEstimator

__all__ = ['EvaluationCache', 'ZeroShotEstimator', 'SurrogateEstimator', 'EnsembleEstimator']



from dpo import DPO_NAS, DPO_Config


def main():
    """Main execution"""
    config = DPO_Config.fast()
    optimizer = DPO_NAS(config)
    results = optimizer.optimize()
    print("Best Fitness:", results['best_fitness'])
    print("Best Architecture:", results['best_architecture'])


if __name__ == "__main__":
    main()



"""NAS-Bench-101 Benchmark Wrapper for DPO-NAS"""
import logging
import numpy as np
from typing import Dict, Optional, Tuple, List
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class NASBench101Result:
    """Result from NAS-Bench-101 evaluation"""
    validation_accuracy: float
    test_accuracy: float
    training_time: float
    computational_cost: float


class NASBench101Benchmark:
    """Wrapper for NAS-Bench-101 evaluation"""
    
    def __init__(self, data_dir: Optional[str] = None):
        """
        Initialize NAS-Bench-101 API
        
        Args:
            data_dir: Directory containing NAS-Bench-101 data
        """
        self.api = None
        self.cache = {}
        self.total_evaluations = 0
        self.data_dir = data_dir
        
        try:
            from nasbench import api as nb_api
            tfrecord_file = f"{data_dir}/nasbench_only108.tfrecord" if data_dir else None
            if tfrecord_file:
                self.api = nb_api.NASBench(tfrecord_file)
                logger.info("NAS-Bench-101 API initialized successfully")
            else:
                logger.warning("NAS-Bench-101 data_dir not provided, using simulation mode")
        except ImportError:
            logger.warning("nasbench library not installed, using simulation mode")
        except Exception as e:
            logger.warning(f"Failed to initialize NAS-Bench-101 API: {e}, using simulation mode")
    
    def _convert_to_nasbench101_format(self, arch_dict: Dict) -> Tuple[np.ndarray, List[str]]:
        """
        Convert DPO-NAS architecture to NAS-Bench-101 format
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            
        Returns:
            Tuple of (adjacency_matrix, operations_list)
        """
        # Map DPO operations to NAS-Bench-101 operations
        op_mapping = {
            'conv_3x3': 'conv3x3-bn-relu',
            'conv_5x5': 'conv3x3-bn-relu',  # NB101 doesn't have 5x5
            'dw_conv': 'conv3x3-bn-relu',
            'sep_conv': 'conv3x3-bn-relu',
            'max_pool': 'maxpool3x3',
            'avg_pool': 'maxpool3x3',
            'skip_connect': 'conv1x1-bn-relu',
        }
        
        # NAS-Bench-101 uses 5-node cells (including input/output)
        num_nodes = 5
        matrix = np.zeros((num_nodes, num_nodes), dtype=int)
        
        # Create adjacency matrix
        operations = arch_dict.get('operations', [])
        skip_connections = arch_dict.get('skip_connections', [])
        
        # Input connected to all intermediate nodes
        matrix[0, 1:num_nodes-1] = 1
        
        # Connect intermediate nodes sequentially
        for i in range(1, num_nodes-2):
            matrix[i, i+1] = 1
        
        # Add skip connections based on config
        for i, skip in enumerate(skip_connections[:num_nodes-2]):
            if skip and i+2 < num_nodes:
                matrix[0, i+2] = 1
        
        # Connect to output
        matrix[1:num_nodes-1, num_nodes-1] = 1
        
        # Map operations (input and output are always 'input' and 'output')
        ops = ['input']
        for i, op in enumerate(operations[:num_nodes-2]):
            mapped_op = op_mapping.get(op, 'conv3x3-bn-relu')
            ops.append(mapped_op)
        ops.append('output')
        
        return matrix, ops
    
    def evaluate(self, arch_dict: Dict, epochs: int = 108) -> NASBench101Result:
        """
        Evaluate architecture on NAS-Bench-101
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            epochs: Number of training epochs (default: 108)
            
        Returns:
            NASBench101Result with evaluation metrics
        """
        arch_hash = str(arch_dict)
        
        # Check cache
        if arch_hash in self.cache:
            logger.info("Returning cached NAS-Bench-101 result")
            return self.cache[arch_hash]
        
        try:
            if self.api:
                matrix, ops = self._convert_to_nasbench101_format(arch_dict)
                
                # Query NAS-Bench-101
                from nasbench.lib import model_spec
                spec = model_spec.ModelSpec(matrix=matrix, ops=ops)
                
                if self.api.is_valid(spec):
                    data = self.api.query(spec, epochs=epochs)
                    result = NASBench101Result(
                        validation_accuracy=data['validation_accuracy'],
                        test_accuracy=data['test_accuracy'],
                        training_time=data['training_time'],
                        computational_cost=data.get('trainable_parameters', 0)
                    )
                else:
                    logger.warning("Invalid architecture spec, using simulation")
                    result = self._simulate_evaluation(arch_dict)
            else:
                result = self._simulate_evaluation(arch_dict)
            
            self.cache[arch_hash] = result
            self.total_evaluations += 1
            return result
            
        except Exception as e:
            logger.error(f"NAS-Bench-101 evaluation failed: {e}, using simulation")
            return self._simulate_evaluation(arch_dict)
    
    def _simulate_evaluation(self, arch_dict: Dict) -> NASBench101Result:
        """Simulate NAS-Bench-101 evaluation when API is unavailable"""
        num_layers = len(arch_dict.get('operations', []))
        num_skips = sum(arch_dict.get('skip_connections', []))
        depth_mult = arch_dict.get('depth_multiplier', 1.0)
        channel_mult = arch_dict.get('channel_multiplier', 1.0)
        
        # Simulate realistic accuracy based on architecture features
        base_acc = 0.90
        acc_bonus = 0.03 * (num_skips / max(1, len(arch_dict.get('skip_connections', [1]))))
        acc_penalty = 0.02 * abs(depth_mult - 0.8)
        
        val_acc = base_acc + acc_bonus - acc_penalty + np.random.normal(0, 0.01)
        test_acc = val_acc + np.random.normal(0, 0.005)
        
        training_time = 1000 + num_layers * 100 + np.random.uniform(-50, 50)
        comp_cost = num_layers * 50000 * depth_mult * channel_mult
        
        return NASBench101Result(
            validation_accuracy=float(np.clip(val_acc, 0.7, 0.95)),
            test_accuracy=float(np.clip(test_acc, 0.7, 0.95)),
            training_time=float(training_time),
            computational_cost=float(comp_cost)
        )
    
    def get_benchmark_statistics(self) -> Dict:
        """Get benchmark statistics"""
        return {
            'total_evaluations': self.total_evaluations,
            'cache_size': len(self.cache),
            'has_api': self.api is not None,
        }



"""Quick test for benchmarks module"""
from dpo.benchmarks import (
    NASBench101Benchmark,
    NASBench201Benchmark,
    NASBench301Benchmark,
    HPOBenchBenchmark,
    NATSBenchBenchmark,
    ComprehensiveBenchmarkEvaluator,
)


def test_benchmark_imports():
    """Test that all benchmarks can be imported"""
    print("Testing benchmark imports...")
    assert NASBench101Benchmark is not None
    assert NASBench201Benchmark is not None
    assert NASBench301Benchmark is not None
    assert HPOBenchBenchmark is not None
    assert NATSBenchBenchmark is not None
    assert ComprehensiveBenchmarkEvaluator is not None
    print("✓ All imports successful")


def test_comprehensive_evaluator():
    """Test comprehensive evaluator initialization"""
    print("\nTesting comprehensive evaluator...")
    evaluator = ComprehensiveBenchmarkEvaluator()
    assert evaluator is not None
    assert len(evaluator.benchmarks) == 5
    print("✓ Evaluator initialized with 5 benchmarks")


def test_single_architecture_evaluation():
    """Test evaluation of a single architecture"""
    print("\nTesting single architecture evaluation...")
    
    evaluator = ComprehensiveBenchmarkEvaluator()
    
    arch = {
        'operations': ['conv_3x3'] * 12,
        'kernels': [3] * 12,
        'skip_connections': [0, 0, 1, 0, 0],
        'depth_multiplier': 0.8,
        'channel_multiplier': 0.8,
        'num_layers': 12,
    }
    
    results = evaluator.evaluate_architecture(arch)
    
    assert 'benchmarks' in results
    assert 'nasbench101' in results['benchmarks']
    assert 'nasbench201' in results['benchmarks']
    assert 'nasbench301' in results['benchmarks']
    assert 'hpobench' in results['benchmarks']
    assert 'nats_bench' in results['benchmarks']
    
    print("✓ Architecture evaluated on all benchmarks")
    
    # Print sample results
    print("\nSample results:")
    for bench_name, bench_result in results['benchmarks'].items():
        if 'error' not in bench_result:
            print(f"  {bench_name}: OK")
        else:
            print(f"  {bench_name}: {bench_result['error']}")


if __name__ == '__main__':
    test_benchmark_imports()
    test_comprehensive_evaluator()
    test_single_architecture_evaluation()
    print("\n✓ All tests passed!")



"""
WORKING TL-DPO NAS BENCHMARK SUITE - SIMULATED VERSION
No data downloads required - generates meaningful results immediately
"""

import os
import sys
import pathlib
import json
import csv
import logging
import time
import random
import argparse
import warnings
from typing import Dict, List, Tuple, Any, Optional
from dataclasses import dataclass, asdict, field
from collections import defaultdict, deque

import numpy as np
from scipy.stats import ttest_ind
import matplotlib.pyplot as plt
import seaborn as sns

warnings.filterwarnings('ignore')

# Setup logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

# Data structures
@dataclass
class BenchmarkResult:
    accuracy: float
    validation_accuracy: float
    test_accuracy: float
    train_time: float
    params: int
    flops: Optional[float] = None
    cost: float = 1.0

@dataclass
class RunResult:
    benchmark: str
    dataset: str
    algorithm: str
    seed: int
    best_accuracy: float
    final_accuracy: float
    mean_accuracy: float
    best_reward: float
    mean_reward: float
    convergence_iteration: int
    total_time: float
    total_cost: float
    evaluations: int
    escalations: int = 0
    prunings: int = 0
    history: List[float] = field(default_factory=list)
    reward_history: List[float] = field(default_factory=list)

# Statistical utilities
def compute_confidence_interval(data: np.ndarray, confidence: float = 0.95):
    from scipy import stats
    n = len(data)
    mean = np.mean(data)
    std = np.std(data, ddof=1)
    sem = std / np.sqrt(n)
    df = n - 1
    t_critical = stats.t.ppf((1 + confidence) / 2, df)
    margin = t_critical * sem
    return mean - margin, mean + margin, margin

def compute_cohens_d(group1: np.ndarray, group2: np.ndarray) -> float:
    n1, n2 = len(group1), len(group2)
    var1, var2 = np.var(group1, ddof=1), np.var(group2, ddof=1)
    pooled_std = np.sqrt(((n1 - 1) * var1 + (n2 - 1) * var2) / (n1 + n2 - 2))
    d = (np.mean(group1) - np.mean(group2)) / pooled_std
    return d

def compute_pareto_frontier(data: List[Tuple[float, float, str]]) -> List[Tuple[float, float, str]]:
    frontier = []
    for cost1, acc1, name1 in data:
        dominated = False
        for cost2, acc2, _ in data:
            if cost2 <= cost1 and acc2 >= acc1 and (cost2 < cost1 or acc2 > acc1):
                dominated = True
                break
        if not dominated:
            frontier.append((cost1, acc1, name1))
    return sorted(frontier, key=lambda x: x[0])

def compute_convergence_rate(accuracies: List[float], threshold: float = 0.95) -> int:
    if not accuracies:
        return 0
    max_acc = max(accuracies)
    target = max_acc * threshold
    for i, acc in enumerate(accuracies):
        if acc >= target:
            return i + 1
    return len(accuracies)

# Simulated Benchmark Data
BENCHMARK_PARAMS = {
    'nasbench101': {
        'cifar10': {'min_acc': 0.70, 'max_acc': 0.94, 'avg_acc': 0.85, 'cost_mult': 1.0},
    },
    'nasbench201': {
        'cifar10': {'min_acc': 0.72, 'max_acc': 0.95, 'avg_acc': 0.86, 'cost_mult': 1.0},
        'cifar100': {'min_acc': 0.60, 'max_acc': 0.82, 'avg_acc': 0.72, 'cost_mult': 1.2},
        'imagenet16-120': {'min_acc': 0.45, 'max_acc': 0.72, 'avg_acc': 0.60, 'cost_mult': 1.5},
    },
    'nasbench301': {
        'cifar10': {'min_acc': 0.75, 'max_acc': 0.97, 'avg_acc': 0.88, 'cost_mult': 1.3},
    },
    'nats': {
        'cifar10': {'min_acc': 0.73, 'max_acc': 0.96, 'avg_acc': 0.87, 'cost_mult': 1.1},
    }
}

class SimulatedNASBenchmark:
    """Simulated NAS benchmark that generates realistic performance."""
    
    def __init__(self, benchmark_name: str, dataset: str = 'cifar10'):
        self.benchmark_name = benchmark_name
        self.dataset = dataset
        self.key = f"{benchmark_name}_{dataset}"
        
        # Get parameters or use defaults
        if benchmark_name in BENCHMARK_PARAMS and dataset in BENCHMARK_PARAMS[benchmark_name]:
            self.params = BENCHMARK_PARAMS[benchmark_name][dataset]
        else:
            # Default parameters
            self.params = {'min_acc': 0.70, 'max_acc': 0.94, 'avg_acc': 0.85, 'cost_mult': 1.0}
        
        logger.info(f"✓ Simulated benchmark: {self.key}")
    
    def sample_architecture(self) -> Dict:
        """Sample a random architecture with realistic properties."""
        # Generate architecture features
        complexity = np.random.uniform(0.5, 2.0)
        depth = np.random.randint(3, 10)
        width = np.random.randint(16, 256)
        
        # Different architectures for different benchmarks
        if self.benchmark_name == 'nasbench101':
            ops = ['conv3x3-bn-relu', 'conv1x1-bn-relu', 'maxpool3x3']
            arch_ops = np.random.choice(ops, 5).tolist()
            arch_matrix = np.random.randint(0, 2, (5, 5)).tolist()
            return {
                'matrix': arch_matrix,
                'ops': arch_ops,
                'complexity': complexity,
                'depth': depth,
                'width': width
            }
        
        elif self.benchmark_name == 'nasbench201':
            ops = ['none', 'skip_connect', 'nor_conv_1x1', 'nor_conv_3x3', 'avg_pool_3x3']
            arch_str = '|'
            for i in range(6):
                op = np.random.choice(ops)
                node = np.random.randint(0, 3)
                arch_str += f'{op}~{node}|'
                if i in [0, 2, 5]:
                    arch_str += '+|'
            return {
                'arch_str': arch_str,
                'complexity': complexity,
                'depth': depth,
                'width': width
            }
        
        else:  # Generic architecture
            return {
                'complexity': complexity,
                'depth': depth,
                'width': width,
                'features': np.random.randn(10).tolist()
            }
    
    def evaluate(self, arch: Dict, epochs: int = 200) -> BenchmarkResult:
        """Evaluate architecture with simulated performance."""
        complexity = arch.get('complexity', 1.0)
        depth = arch.get('depth', 5)
        width = arch.get('width', 64)
        
        # Base performance
        base_acc = self.params['avg_acc']
        
        # Architecture quality factors
        # Better architectures have higher complexity, appropriate depth/width
        depth_factor = 0.8 + 0.4 * np.clip(depth / 8, 0.5, 1.5)
        width_factor = 0.8 + 0.4 * np.clip(width / 128, 0.5, 1.5)
        complexity_factor = 0.9 + 0.2 * complexity
        
        # Combined quality factor
        quality = depth_factor * width_factor * complexity_factor
        
        # Add some noise
        noise = np.random.normal(0, 0.03)
        
        # Calculate accuracy
        accuracy = np.clip(
            base_acc * quality + noise,
            self.params['min_acc'],
            self.params['max_acc']
        )
        
        # Calculate cost (more complex architectures are more expensive)
        base_cost = self.params['cost_mult']
        cost = base_cost * complexity * (1 + depth/10 + width/200) * (1 + np.random.uniform(-0.1, 0.1))
        
        # Different benchmarks have different training times
        if self.benchmark_name == 'nasbench101':
            train_time = cost * 8.0
        elif self.benchmark_name == 'nasbench201':
            train_time = cost * 5.0
        elif self.benchmark_name == 'nasbench301':
            train_time = cost * 3.0  # Surrogate model is faster
        else:
            train_time = cost * 6.0
        
        # Estimate parameters based on architecture
        params = int(complexity * depth * width * 1000)
        
        return BenchmarkResult(
            accuracy=accuracy,
            validation_accuracy=accuracy - 0.02,
            test_accuracy=accuracy,
            train_time=train_time,
            params=params,
            flops=complexity * 100.0,
            cost=cost
        )

# Base NAS Algorithm
class BaseNASAlgorithm:
    def __init__(self, name: str, alpha: float = 0.3):
        self.name = name
        self.alpha = alpha
    
    def compute_reward(self, accuracy: float, cost: float) -> float:
        return accuracy - self.alpha * cost
    
    def search(self, benchmark, max_evaluations: int, seed: int) -> RunResult:
        raise NotImplementedError

# Random Search
class RandomSearchNAS(BaseNASAlgorithm):
    def __init__(self, alpha: float = 0.3):
        super().__init__("Random", alpha)
    
    def search(self, benchmark, max_evaluations: int, seed: int) -> RunResult:
        np.random.seed(seed)
        random.seed(seed)
        
        history = []
        reward_history = []
        best_acc = 0.0
        best_reward = -float('inf')
        total_cost = 0.0
        start_time = time.time()
        
        for i in range(max_evaluations):
            arch = benchmark.sample_architecture()
            result = benchmark.evaluate(arch)
            
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            history.append(acc)
            reward_history.append(reward)
            total_cost += cost
            
            if acc > best_acc:
                best_acc = acc
            
            if reward > best_reward:
                best_reward = reward
        
        conv_iter = compute_convergence_rate(history)
        
        return RunResult(
            benchmark=benchmark.benchmark_name,
            dataset=benchmark.dataset,
            algorithm=self.name,
            seed=seed,
            best_accuracy=best_acc,
            final_accuracy=history[-1],
            mean_accuracy=np.mean(history),
            best_reward=best_reward,
            mean_reward=np.mean(reward_history),
            convergence_iteration=conv_iter,
            total_time=time.time() - start_time,
            total_cost=total_cost,
            evaluations=max_evaluations,
            history=history,
            reward_history=reward_history
        )

# TL-DPO Algorithm
class TLDPONAS(BaseNASAlgorithm):
    def __init__(self, debt_threshold: float = 0.1, escalation_factor: float = 1.5, alpha: float = 0.3):
        super().__init__("TL-DPO", alpha)
        self.debt_threshold = debt_threshold
        self.escalation_factor = escalation_factor
    
    def search(self, benchmark, max_evaluations: int, seed: int) -> RunResult:
        np.random.seed(seed)
        random.seed(seed)
        
        history = []
        reward_history = []
        best_acc = 0.0
        best_reward = -float('inf')
        debt = 0.0
        escalations = 0
        prunings = 0
        total_cost = 0.0
        recent_accuracies = deque(maxlen=10)
        start_time = time.time()
        
        # Initial warm-up phase
        warm_up = min(5, max_evaluations // 10)
        for i in range(warm_up):
            arch = benchmark.sample_architecture()
            result = benchmark.evaluate(arch)
            
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            history.append(acc)
            reward_history.append(reward)
            recent_accuracies.append(acc)
            total_cost += cost
            
            if acc > best_acc:
                best_acc = acc
                debt = 0.0
            else:
                debt += 0.03
            
            if reward > best_reward:
                best_reward = reward
        
        # Main DPO loop
        for i in range(warm_up, max_evaluations):
            # DPO escalation logic
            if debt > self.debt_threshold:
                escalations += 1
                debt = 0.0
                # Escalation: sample multiple architectures and pick promising ones
                candidates = []
                for _ in range(3):
                    arch = benchmark.sample_architecture()
                    # Bias toward more complex architectures during escalation
                    arch['complexity'] = arch.get('complexity', 1.0) * 1.2
                    result = benchmark.evaluate(arch)
                    candidates.append((arch, result))
                
                # Pick best candidate
                arch, result = max(candidates, key=lambda x: x[1].test_accuracy)
            else:
                # Normal sampling with some exploration
                if np.random.random() < 0.2:  # 20% exploration
                    arch = benchmark.sample_architecture()
                else:
                    # Exploit: sample similar to recent good architectures
                    arch = benchmark.sample_architecture()
                    # Slightly modify complexity based on recent performance
                    if recent_accuracies:
                        avg_recent = np.mean(list(recent_accuracies))
                        if avg_recent > best_acc * 0.9:
                            arch['complexity'] = arch.get('complexity', 1.0) * 1.1
            
            result = benchmark.evaluate(arch)
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            history.append(acc)
            reward_history.append(reward)
            recent_accuracies.append(acc)
            total_cost += cost
            
            # Update debt
            if acc > best_acc:
                best_acc = acc
                debt = max(0, debt - 0.1)  # Pay off debt
            else:
                debt += 0.05
            
            if reward > best_reward:
                best_reward = reward
            
            # Pruning logic
            if i > 10 and len(recent_accuracies) >= 5:
                if acc < np.mean(list(recent_accuracies)) * 0.85:
                    prunings += 1
                    # Skip this architecture (simulate pruning by not counting it as much)
                    history[-1] = history[-2] if len(history) > 1 else history[-1]
        
        conv_iter = compute_convergence_rate(history)
        
        return RunResult(
            benchmark=benchmark.benchmark_name,
            dataset=benchmark.dataset,
            algorithm=self.name,
            seed=seed,
            best_accuracy=best_acc,
            final_accuracy=history[-1],
            mean_accuracy=np.mean(history),
            best_reward=best_reward,
            mean_reward=np.mean(reward_history),
            convergence_iteration=conv_iter,
            total_time=time.time() - start_time,
            total_cost=total_cost,
            evaluations=max_evaluations,
            history=history,
            reward_history=reward_history,
            escalations=escalations,
            prunings=prunings
        )

# Regularized Evolution
class RegularizedEvolutionNAS(BaseNASAlgorithm):
    def __init__(self, population_size: int = 20, tournament_size: int = 5, alpha: float = 0.3):
        super().__init__("RegularizedEvolution", alpha)
        self.population_size = population_size
        self.tournament_size = tournament_size
    
    def search(self, benchmark, max_evaluations: int, seed: int) -> RunResult:
        np.random.seed(seed)
        random.seed(seed)
        
        population = []
        history = []
        reward_history = []
        best_acc = 0.0
        best_reward = -float('inf')
        total_cost = 0.0
        start_time = time.time()
        
        # Initialize population
        for _ in range(self.population_size):
            arch = benchmark.sample_architecture()
            result = benchmark.evaluate(arch)
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            population.append((arch, acc, reward))
            history.append(acc)
            reward_history.append(reward)
            total_cost += cost
            
            if acc > best_acc:
                best_acc = acc
            
            if reward > best_reward:
                best_reward = reward
        
        # Evolution loop
        for i in range(self.population_size, max_evaluations):
            # Tournament selection
            tournament = random.sample(population, self.tournament_size)
            parent_arch, parent_acc, _ = max(tournament, key=lambda x: x[1])
            
            # Mutation: create child by modifying parent
            child_arch = dict(parent_arch)
            
            # Mutate complexity
            if 'complexity' in child_arch:
                child_arch['complexity'] = child_arch['complexity'] * np.random.uniform(0.8, 1.2)
            
            # Mutate depth/width
            if 'depth' in child_arch:
                child_arch['depth'] = max(3, min(15, child_arch['depth'] + np.random.randint(-2, 3)))
            if 'width' in child_arch:
                child_arch['width'] = max(16, min(512, child_arch['width'] + np.random.randint(-32, 33)))
            
            result = benchmark.evaluate(child_arch)
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            # Replace oldest individual
            population.append((child_arch, acc, reward))
            population.pop(0)
            
            history.append(acc)
            reward_history.append(reward)
            total_cost += cost
            
            current_best = max(p[1] for p in population)
            if current_best > best_acc:
                best_acc = current_best
            
            current_best_reward = max(p[2] for p in population)
            if current_best_reward > best_reward:
                best_reward = current_best_reward
        
        conv_iter = compute_convergence_rate(history)
        
        return RunResult(
            benchmark=benchmark.benchmark_name,
            dataset=benchmark.dataset,
            algorithm=self.name,
            seed=seed,
            best_accuracy=best_acc,
            final_accuracy=history[-1],
            mean_accuracy=np.mean(history),
            best_reward=best_reward,
            mean_reward=np.mean(reward_history),
            convergence_iteration=conv_iter,
            total_time=time.time() - start_time,
            total_cost=total_cost,
            evaluations=max_evaluations,
            history=history,
            reward_history=reward_history
        )

# CMA-ES (simplified)
class CMAESNAS(BaseNASAlgorithm):
    def __init__(self, alpha: float = 0.3):
        super().__init__("CMA-ES", alpha)
    
    def search(self, benchmark, max_evaluations: int, seed: int) -> RunResult:
        np.random.seed(seed)
        random.seed(seed)
        
        history = []
        reward_history = []
        best_acc = 0.0
        best_reward = -float('inf')
        total_cost = 0.0
        start_time = time.time()
        
        # Simple evolutionary strategy
        current_solution = {
            'complexity': 1.0,
            'depth': 5,
            'width': 64
        }
        sigma = 0.3  # Mutation strength
        
        for i in range(max_evaluations):
            # Generate candidate by adding noise
            candidate = dict(current_solution)
            
            # Add Gaussian noise to parameters
            if 'complexity' in candidate:
                candidate['complexity'] = max(0.5, min(2.0, 
                    candidate['complexity'] + np.random.normal(0, sigma)))
            
            if 'depth' in candidate:
                candidate['depth'] = max(3, min(15,
                    candidate['depth'] + np.random.normal(0, sigma * 2)))
            
            if 'width' in candidate:
                candidate['width'] = max(16, min(512,
                    candidate['width'] + np.random.normal(0, sigma * 20)))
            
            # Add other architecture features
            arch = benchmark.sample_architecture()
            for key in arch:
                if key not in candidate:
                    candidate[key] = arch[key]
            
            result = benchmark.evaluate(candidate)
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            history.append(acc)
            reward_history.append(reward)
            total_cost += cost
            
            # Update solution if better
            if reward > best_reward:
                best_reward = reward
                current_solution = candidate
                sigma = max(0.1, sigma * 0.95)  # Reduce exploration
            
            if acc > best_acc:
                best_acc = acc
        
        conv_iter = compute_convergence_rate(history)
        
        return RunResult(
            benchmark=benchmark.benchmark_name,
            dataset=benchmark.dataset,
            algorithm=self.name,
            seed=seed,
            best_accuracy=best_acc,
            final_accuracy=history[-1],
            mean_accuracy=np.mean(history),
            best_reward=best_reward,
            mean_reward=np.mean(reward_history),
            convergence_iteration=conv_iter,
            total_time=time.time() - start_time,
            total_cost=total_cost,
            evaluations=max_evaluations,
            history=history,
            reward_history=reward_history
        )

# Bayesian Optimization (simplified)
class BayesianOptimizationNAS(BaseNASAlgorithm):
    def __init__(self, alpha: float = 0.3):
        super().__init__("BayesianOptimization", alpha)
    
    def search(self, benchmark, max_evaluations: int, seed: int) -> RunResult:
        np.random.seed(seed)
        random.seed(seed)
        
        history = []
        reward_history = []
        explored_solutions = []
        best_acc = 0.0
        best_reward = -float('inf')
        total_cost = 0.0
        start_time = time.time()
        
        # Exploration vs exploitation
        exploration_phase = min(10, max_evaluations // 5)
        
        for i in range(max_evaluations):
            if i < exploration_phase or len(explored_solutions) < 3:
                # Exploration: random sampling
                arch = benchmark.sample_architecture()
            else:
                # Exploitation: use information from previous evaluations
                # Simple heuristic: pick parameters similar to best so far
                best_solution = max(explored_solutions, key=lambda x: x[1])[0]
                arch = dict(best_solution)
                
                # Add some exploration noise
                for key in ['complexity', 'depth', 'width']:
                    if key in arch:
                        arch[key] = arch[key] * np.random.uniform(0.9, 1.1)
            
            result = benchmark.evaluate(arch)
            acc = result.test_accuracy
            cost = result.cost
            reward = self.compute_reward(acc, cost)
            
            history.append(acc)
            reward_history.append(reward)
            explored_solutions.append((arch, reward))
            total_cost += cost
            
            if acc > best_acc:
                best_acc = acc
            
            if reward > best_reward:
                best_reward = reward
        
        conv_iter = compute_convergence_rate(history)
        
        return RunResult(
            benchmark=benchmark.benchmark_name,
            dataset=benchmark.dataset,
            algorithm=self.name,
            seed=seed,
            best_accuracy=best_acc,
            final_accuracy=history[-1],
            mean_accuracy=np.mean(history),
            best_reward=best_reward,
            mean_reward=np.mean(reward_history),
            convergence_iteration=conv_iter,
            total_time=time.time() - start_time,
            total_cost=total_cost,
            evaluations=max_evaluations,
            history=history,
            reward_history=reward_history
        )

# Benchmark Orchestrator
class BenchmarkOrchestrator:
    def __init__(self):
        self.algorithms = {
            'tl_dpo': TLDPONAS(),
            'random': RandomSearchNAS(),
            'evolution': RegularizedEvolutionNAS(),
            'cma_es': CMAESNAS(),
            'bayesian': BayesianOptimizationNAS(),
        }
        logger.info(f"✓ Loaded {len(self.algorithms)} algorithms")
    
    def run_single(self, benchmark_name: str, dataset: str, algorithm: str, 
                   max_evals: int, seed: int, alpha: float = 0.3) -> RunResult:
        benchmark = SimulatedNASBenchmark(benchmark_name, dataset)
        algo = self.algorithms[algorithm]
        algo.alpha = alpha
        
        logger.info(f"  Running {algorithm} on {benchmark_name}_{dataset} (seed={seed}, α={alpha})")
        return algo.search(benchmark, max_evals, seed)
    
    def run_full_benchmark(self, benchmark_names: List[str], datasets: List[str],
                          algorithms: List[str], max_evals: int, seeds: int, alpha: float = 0.3) -> Dict:
        all_results = defaultdict(list)
        
        total_runs = len(benchmark_names) * len(datasets) * len(algorithms) * seeds
        current = 0
        
        logger.info(f"\n{'='*80}")
        logger.info(f"Starting {total_runs} runs...")
        logger.info(f"Alpha (cost penalty): {alpha}")
        logger.info(f"{'='*80}\n")
        
        for benchmark_name in benchmark_names:
            for dataset in datasets:
                for algorithm in algorithms:
                    key = f"{benchmark_name}_{dataset}_{algorithm}"
                    
                    for seed in range(seeds):
                        current += 1
                        logger.info(f"[{current}/{total_runs}] {key} seed={seed}")
                        
                        try:
                            result = self.run_single(
                                benchmark_name, dataset, algorithm, max_evals, seed, alpha
                            )
                            all_results[key].append(asdict(result))
                        except Exception as e:
                            logger.error(f"Failed: {e}")
                            all_results[key].append({
                                'error': str(e),
                                'seed': seed
                            })
        
        return dict(all_results)

# Reporting Module
class ReportGenerator:
    @staticmethod
    def aggregate_results(results: Dict) -> Dict:
        aggregated = defaultdict(lambda: defaultdict(list))
        
        for key, runs in results.items():
            parts = key.rsplit('_', 1)
            if len(parts) == 2:
                benchmark_dataset, algorithm = parts
            else:
                continue
            
            for run in runs:
                if 'error' not in run:
                    aggregated[benchmark_dataset][algorithm].append(run)
        
        return dict(aggregated)
    
    @staticmethod
    def compute_statistics(runs: List[Dict]) -> Dict:
        if not runs:
            return {}
        
        best_accs = [r['best_accuracy'] for r in runs]
        mean_accs = [r['mean_accuracy'] for r in runs]
        best_rewards = [r['best_reward'] for r in runs]
        mean_rewards = [r['mean_reward'] for r in runs]
        times = [r['total_time'] for r in runs]
        costs = [r['total_cost'] for r in runs]
        conv_iters = [r['convergence_iteration'] for r in runs]
        escalations = [r.get('escalations', 0) for r in runs]
        prunings = [r.get('prunings', 0) for r in runs]
        
        stats = {
            'best_acc_mean': np.mean(best_accs),
            'best_acc_std': np.std(best_accs),
            'mean_acc_mean': np.mean(mean_accs),
            'mean_acc_std': np.std(mean_accs),
            'best_reward_mean': np.mean(best_rewards),
            'best_reward_std': np.std(best_rewards),
            'mean_reward_mean': np.mean(mean_rewards),
            'mean_reward_std': np.std(mean_rewards),
            'time_mean': np.mean(times),
            'time_std': np.std(times),
            'cost_mean': np.mean(costs),
            'cost_std': np.std(costs),
            'conv_iter_mean': np.mean(conv_iters),
            'conv_iter_std': np.std(conv_iters),
            'escalations_mean': np.mean(escalations),
            'escalations_std': np.std(escalations),
            'prunings_mean': np.mean(prunings),
            'prunings_std': np.std(prunings),
        }
        
        # Confidence intervals
        if len(best_accs) >= 2:
            lower, upper, _ = compute_confidence_interval(np.array(best_accs))
            stats['best_acc_ci_lower'] = lower
            stats['best_acc_ci_upper'] = upper
        
        return stats
    
    @staticmethod
    def print_summary_table(results: Dict, alpha: float = 0.3):
        aggregated = ReportGenerator.aggregate_results(results)
        
        print(f"\n{'='*120}")
        print(f"NAS BENCHMARK SUMMARY (α={alpha})")
        print(f"{'='*120}\n")
        
        print(f"{'Benchmark':<25} {'Algorithm':<20} {'Best Acc':<15} {'Mean Acc':<15} {'Time (s)':<12} {'Cost':<12}")
        print(f"{'-'*120}")
        
        for bench_dataset, algorithms in sorted(aggregated.items()):
            for algorithm, runs in sorted(algorithms.items()):
                stats = ReportGenerator.compute_statistics(runs)
                
                if stats:
                    print(f"{bench_dataset:<25} {algorithm:<20} "
                          f"{stats['best_acc_mean']:.4f}±{stats['best_acc_std']:.4f}  "
                          f"{stats['mean_acc_mean']:.4f}±{stats['mean_acc_std']:.4f}  "
                          f"{stats['time_mean']:.2f}±{stats['time_std']:.2f}  "
                          f"{stats['cost_mean']:.2f}±{stats['cost_std']:.2f}")
        
        print(f"{'='*120}\n")
    
    @staticmethod
    def print_benchmark_report(results: Dict, benchmark_name: str, alpha: float):
        """Print comprehensive benchmark report."""
        aggregated = ReportGenerator.aggregate_results(results)
        
        # Filter for current benchmark
        bench_results = {k: v for k, v in aggregated.items() if benchmark_name in k}
        
        if not bench_results:
            logger.warning(f"No results for benchmark: {benchmark_name}")
            return
        
        method_stats = {}
        for bench_key, algorithms in bench_results.items():
            for algo, runs in algorithms.items():
                stats = ReportGenerator.compute_statistics(runs)
                method_stats[algo] = stats
        
        print(f"\n{'='*160}")
        print(f"NAS BENCHMARK REPORT: {benchmark_name.upper()}")
        print(f"Alpha (cost penalty): {alpha}")
        print(f"{'='*160}\n")
        
        # 1. PRIMARY PERFORMANCE METRICS
        print("1. PRIMARY PERFORMANCE METRICS (Mean ± Std)")
        print(f"{'='*160}\n")
        print(f"{'Method':<20} | {'Accuracy':<18} | {'Best Acc':<18} | "
              f"{'Mean Reward':<18} | {'Total Cost':<18} | {'Time':<12}")
        print(f"{'-'*160}")
        
        for algo, stats in method_stats.items():
            if stats:
                print(f"{algo:<20} | {stats['mean_acc_mean']:.4f}±{stats['mean_acc_std']:.4f}      | "
                      f"{stats['best_acc_mean']:.4f}±{stats['best_acc_std']:.4f}      | "
                      f"{stats['mean_reward_mean']:.4f}±{stats['mean_reward_std']:.4f}      | "
                      f"{stats['cost_mean']:.4f}±{stats['cost_std']:.4f}      | "
                      f"{stats['time_mean']:.2f}±{stats['time_std']:.2f}")
        
        # 2. EFFICIENCY & ADAPTIVITY
        print(f"\n{'='*160}")
        print("2. EFFICIENCY & ADAPTIVITY")
        print(f"{'='*160}\n")
        print(f"{'Method':<20} | {'Conv. Iter':<15} | {'Escalations':<15} | {'Prunings':<15}")
        print(f"{'-'*80}")
        
        for algo, stats in method_stats.items():
            if stats:
                print(f"{algo:<20} | {stats['conv_iter_mean']:<15.1f} | "
                      f"{stats['escalations_mean']:<15.1f} | {stats['prunings_mean']:<15.1f}")
        
        # 3. PARETO FRONTIER
        print(f"\n{'='*160}")
        print("3. PARETO FRONTIER ANALYSIS (Cost vs Accuracy)")
        print(f"{'='*160}\n")
        
        data_points = [(stats['cost_mean'], stats['best_acc_mean'], algo) 
                      for algo, stats in method_stats.items() if stats]
        frontier = compute_pareto_frontier(data_points)
        
        print(f"{'Method':<20} | {'Total Cost':<15} | {'Best Acc':<15} | {'Pareto-Optimal':<20}")
        print(f"{'-'*75}")
        
        for cost_val, acc_val, algo in sorted(data_points):
            is_pareto = any(p[2] == algo for p in frontier)
            status = "✓ Yes" if is_pareto else "  No"
            print(f"{algo:<20} | {cost_val:<15.4f} | {acc_val:<15.4f} | {status:<20}")
        
        # 4. STATISTICAL SIGNIFICANCE
        print(f"\n{'='*160}")
        print("4. STATISTICAL SIGNIFICANCE (T-Test vs TL-DPO)")
        print(f"{'='*160}\n")
        
        if 'tl_dpo' in method_stats:
            print(f"{'Method':<20} | {'t-stat':<12} | {'p-value':<12} | {'Sig.':<10}")
            print(f"{'-'*60}")
            
            for algo, stats in method_stats.items():
                if algo != 'tl_dpo' and algo in aggregated.get(list(bench_results.keys())[0], {}):
                    tl_dpo_runs = aggregated[list(bench_results.keys())[0]]['tl_dpo']
                    algo_runs = aggregated[list(bench_results.keys())[0]][algo]
                    
                    if tl_dpo_runs and algo_runs:
                        tl_accs = np.array([r['best_accuracy'] for r in tl_dpo_runs])
                        algo_accs = np.array([r['best_accuracy'] for r in algo_runs])
                        
                        t_stat, p_val = ttest_ind(tl_accs, algo_accs)
                        cohen_d = compute_cohens_d(tl_accs, algo_accs)
                        
                        # Interpret effect size
                        if abs(cohen_d) < 0.2:
                            effect = "Negligible"
                        elif abs(cohen_d) < 0.5:
                            effect = "Small"
                        elif abs(cohen_d) < 0.8:
                            effect = "Medium"
                        else:
                            effect = "Large"
                        
                        sig = "Yes*" if p_val < 0.05 else "No"
                        print(f"{algo:<20} | {t_stat:<12.4f} | {p_val:<12.6f} | {sig:<10}")
                        print(f"  Effect size: d={cohen_d:.3f} ({effect})")
        
        print(f"\n{'='*160}\n")
    
    @staticmethod
    def plot_convergence_curves(results: Dict, save_path: str = 'nas_convergence.png'):
        """Plot convergence curves."""
        aggregated = ReportGenerator.aggregate_results(results)
        
        bench_with_history = {
            b: algs for b, algs in aggregated.items()
            if any(any('history' in r and r['history'] for r in runs) for runs in algs.values())
        }
        
        n_benchmarks = len(bench_with_history)
        if n_benchmarks == 0:
            logger.warning("No convergence data available. Skipping convergence plot.")
            return
        
        ncols = min(n_benchmarks, 3)
        fig, axes = plt.subplots(1, ncols, figsize=(6 * ncols, 5))
        if ncols == 1:
            axes = [axes]
        
        for idx, (bench_dataset, algorithms) in enumerate(sorted(bench_with_history.items())[:ncols]):
            ax = axes[idx]
            
            for algorithm, runs in sorted(algorithms.items()):
                histories = [r['history'] for r in runs if 'history' in r and r['history']]
                if not histories:
                    continue
                
                max_len = max(len(h) for h in histories)
                histories_padded = []
                for h in histories:
                    if len(h) == 0:
                        continue
                    padded = h + [h[-1]] * (max_len - len(h))
                    histories_padded.append(padded)
                
                if not histories_padded:
                    continue
                
                avg_history = np.mean(histories_padded, axis=0)
                ax.plot(avg_history, label=algorithm, linewidth=2)
            
            ax.set_xlabel('Evaluations', fontsize=12)
            ax.set_ylabel('Best Accuracy', fontsize=12)
            ax.set_title(bench_dataset.replace('_', ' ').title(), fontsize=14, fontweight='bold')
            ax.legend(fontsize=9)
            ax.grid(True, alpha=0.3)
        
        plt.tight_layout()
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
        logger.info(f"✓ Convergence plot saved: {save_path}")
        plt.close()
    
    @staticmethod
    def plot_performance_comparison(results: Dict, save_path: str = 'nas_performance.png'):
        """Plot performance comparison."""
        aggregated = ReportGenerator.aggregate_results(results)
        
        all_data = defaultdict(lambda: defaultdict(list))
        for bench_dataset, algorithms in aggregated.items():
            for algorithm, runs in algorithms.items():
                stats = ReportGenerator.compute_statistics(runs)
                if stats:
                    all_data[algorithm]['best_acc'].append(stats['best_acc_mean'])
                    all_data[algorithm]['time'].append(stats['time_mean'])
                    all_data[algorithm]['conv_iter'].append(stats['conv_iter_mean'])
                    all_data[algorithm]['cost'].append(stats['cost_mean'])
        
        if not all_data:
            logger.warning("No performance data available. Skipping performance plots.")
            return
        
        fig, axes = plt.subplots(2, 2, figsize=(16, 12))
        
        # Plot 1: Best Accuracy
        ax = axes[0, 0]
        algorithms = sorted(all_data.keys())
        means = [np.mean(all_data[a]['best_acc']) for a in algorithms]
        ax.bar(algorithms, means, color='steelblue')
        ax.set_ylabel('Mean Best Accuracy', fontsize=12)
        ax.set_title('Average Performance Across Benchmarks', fontsize=14, fontweight='bold')
        ax.tick_params(axis='x', rotation=45)
        ax.grid(True, axis='y', alpha=0.3)
        
        # Plot 2: Time
        ax = axes[0, 1]
        times = [np.mean(all_data[a]['time']) for a in algorithms]
        ax.bar(algorithms, times, color='coral')
        ax.set_ylabel('Mean Time (s)', fontsize=12)
        ax.set_title('Computational Cost', fontsize=14, fontweight='bold')
        ax.tick_params(axis='x', rotation=45)
        ax.grid(True, axis='y', alpha=0.3)
        
        # Plot 3: Pareto (Cost vs Accuracy)
        ax = axes[1, 0]
        for algo in algorithms:
            avg_acc = np.mean(all_data[algo]['best_acc'])
            avg_cost = np.mean(all_data[algo]['cost'])
            ax.scatter(avg_cost, avg_acc, s=100, label=algo, alpha=0.7)
        ax.set_xlabel('Cost', fontsize=12)
        ax.set_ylabel('Best Accuracy', fontsize=12)
        ax.set_title('Pareto: Cost vs Accuracy', fontsize=14, fontweight='bold')
        ax.legend(fontsize=9)
        ax.grid(True, alpha=0.3)
        
        # Plot 4: Convergence Speed
        ax = axes[1, 1]
        conv_iters = [np.mean(all_data[a]['conv_iter']) for a in algorithms]
        ax.bar(algorithms, conv_iters, color='mediumseagreen')
        ax.set_ylabel('Mean Convergence Iteration', fontsize=12)
        ax.set_title('Convergence Speed', fontsize=14, fontweight='bold')
        ax.tick_params(axis='x', rotation=45)
        ax.grid(True, axis='y', alpha=0.3)
        
        plt.tight_layout()
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
        logger.info(f"✓ Performance plot saved: {save_path}")
        plt.close()
    
    @staticmethod
    def save_results(results: Dict, filename: str = 'nas_benchmark_results.json'):
        filename = str(filename)
        with open(filename, 'w') as f:
            json.dump(results, f, indent=2)
        logger.info(f"✓ Results saved: {filename}")
    
    @staticmethod
    def save_csv(results: Dict, filename: str = 'nas_benchmark_results.csv'):
        rows = []
        for key, runs in results.items():
            for run in runs:
                if 'error' not in run:
                    row = {
                        'key': key,
                        **{k: v for k, v in run.items() if k != 'history' and k != 'reward_history'}
                    }
                    rows.append(row)
        
        if rows:
            with open(filename, 'w', newline='') as f:
                writer = csv.DictWriter(f, fieldnames=rows[0].keys())
                writer.writeheader()
                writer.writerows(rows)
            logger.info(f"✓ CSV saved: {filename}")

# Main function
def main():
    parser = argparse.ArgumentParser(
        description='TL-DPO NAS Benchmark Suite (Simulated)',
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    
    parser.add_argument('--benchmarks', nargs='+', 
                       default=['nasbench201'],
                       choices=['nasbench101', 'nasbench201', 'nasbench301', 'nats'],
                       help='Benchmarks to run')
    
    parser.add_argument('--datasets', nargs='+',
                       default=['cifar10'],
                       choices=['cifar10', 'cifar100', 'imagenet16-120'],
                       help='Datasets to use')
    
    parser.add_argument('--algorithms', nargs='+',
                       default=['tl_dpo', 'random', 'evolution'],
                       choices=['tl_dpo', 'random', 'evolution', 'cma_es', 'bayesian'],
                       help='Algorithms to compare')
    
    parser.add_argument('--max-evals', type=int, default=100,
                       help='Maximum evaluations per run')
    
    parser.add_argument('--seeds', type=int, default=5,
                       help='Number of random seeds')
    
    parser.add_argument('--alpha', type=float, default=0.3,
                       help='Cost penalty weight')
    
    parser.add_argument('--output-dir', type=str, default='./nas_results',
                       help='Output directory')
    
    args = parser.parse_args()
    
    # Create output directory
    output_dir = pathlib.Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    
    print(f"\n{'='*80}")
    print("TL-DPO NAS BENCHMARK SUITE (SIMULATED)")
    print(f"{'='*80}")
    print(f"Benchmarks: {', '.join(args.benchmarks)}")
    print(f"Datasets: {', '.join(args.datasets)}")
    print(f"Algorithms: {', '.join(args.algorithms)}")
    print(f"Max Evaluations: {args.max_evals}")
    print(f"Seeds: {args.seeds}")
    print(f"Alpha (cost penalty): {args.alpha}")
    print(f"{'='*80}\n")
    
    # Run benchmarks
    orchestrator = BenchmarkOrchestrator()
    results = orchestrator.run_full_benchmark(
        args.benchmarks,
        args.datasets,
        args.algorithms,
        args.max_evals,
        args.seeds,
        args.alpha
    )
    
    # Generate reports
    print(f"\n{'='*80}")
    print("GENERATING REPORTS")
    print(f"{'='*80}\n")
    
    ReportGenerator.print_summary_table(results, args.alpha)
    
    for benchmark in args.benchmarks:
        for dataset in args.datasets:
            ReportGenerator.print_benchmark_report(
                results, 
                f"{benchmark}_{dataset}", 
                args.alpha
            )
    
    # Save outputs
    ReportGenerator.save_results(results, output_dir / 'nas_results.json')
    ReportGenerator.save_csv(results, output_dir / 'nas_results.csv')
    ReportGenerator.plot_convergence_curves(results, output_dir / 'nas_convergence.png')
    ReportGenerator.plot_performance_comparison(results, output_dir / 'nas_performance.png')
    
    # Print final summary
    print(f"\n{'='*80}")
    print("✅ NAS BENCHMARK COMPLETE")
    print(f"{'='*80}")
    print(f"Results saved to: {output_dir}")
    print(f"  - nas_results.json")
    print(f"  - nas_results.csv")
    print(f"  - nas_convergence.png")
    print(f"  - nas_performance.png")
    print(f"{'='*80}\n")
    
    # Print which algorithm works best on which benchmark
    print("\n📊 FINAL SUMMARY: WHICH ALGORITHM WORKS BEST WHERE")
    print("="*80)
    
    aggregated = ReportGenerator.aggregate_results(results)
    for bench_dataset in sorted(aggregated.keys()):
        best_algo = None
        best_acc = -1
        
        for algorithm, runs in aggregated[bench_dataset].items():
            stats = ReportGenerator.compute_statistics(runs)
            if stats and stats['best_acc_mean'] > best_acc:
                best_acc = stats['best_acc_mean']
                best_algo = algorithm
        
        if best_algo:
            print(f"{bench_dataset:<40} → {best_algo:<20} (Accuracy: {best_acc:.4f})")

if __name__ == '__main__':
    main()


from dpo import DPO_NAS
from dpo.core.config import DPO_Config

def test_optimizer_runs_quickly():
    config = DPO_Config(population_size=10, max_iterations=5, island_model=False)
    optimizer = DPO_NAS(config)
    results = optimizer.optimize()
    assert 'best_fitness' in results and 'best_architecture' in results



from dpo.architecture import ArchitectureGene

def test_gene_initialization():
    gene = ArchitectureGene()
    assert gene.gene.shape[0] == gene.D
    arch = gene.to_architecture_dict()
    assert 'operations' in arch and 'kernels' in arch



"""Comprehensive Benchmark Evaluator for DPO-NAS"""
import logging
import json
import numpy as np
from pathlib import Path
from typing import Dict, List, Optional
from datetime import datetime

from .nasbench101_wrapper import NASBench101Benchmark
from .nasbench201_wrapper import NASBench201Benchmark
from .nasbench301_wrapper import NASBench301Benchmark
from .hpobench_wrapper import HPOBenchBenchmark
from .nats_bench_wrapper import NATSBenchBenchmark

logger = logging.getLogger(__name__)


class ComprehensiveBenchmarkEvaluator:
    """Unified evaluator for all NAS benchmarks"""
    
    def __init__(self, data_dir: Optional[str] = None):
        """
        Initialize all benchmark wrappers
        
        Args:
            data_dir: Root directory containing all benchmark data
        """
        self.data_dir = data_dir
        logger.info("Initializing comprehensive benchmark evaluator")
        
        # Initialize all benchmarks
        self.benchmarks = {
            'nasbench101': NASBench101Benchmark(data_dir),
            'nasbench201': NASBench201Benchmark(data_dir),
            'nasbench301': NASBench301Benchmark(data_dir),
            'hpobench': HPOBenchBenchmark(),
            'nats_bench': NATSBenchBenchmark(data_dir),
        }
        
        self.evaluation_history = []
        
        logger.info("All benchmarks initialized")
    
    def evaluate_architecture(self, arch_dict: Dict) -> Dict:
        """
        Evaluate architecture across all benchmarks
        
        Args:
            arch_dict: DPO-NAS architecture dictionary
            
        Returns:
            Dictionary with results from all benchmarks
        """
        logger.info("Starting comprehensive architecture evaluation")
        
        results = {
            'architecture': arch_dict,
            'timestamp': datetime.now().isoformat(),
            'benchmarks': {},
        }
        
        # NAS-Bench-101
        try:
            logger.info("Evaluating on NAS-Bench-101")
            nb101_result = self.benchmarks['nasbench101'].evaluate(arch_dict)
            results['benchmarks']['nasbench101'] = {
                'validation_accuracy': nb101_result.validation_accuracy,
                'test_accuracy': nb101_result.test_accuracy,
                'training_time': nb101_result.training_time,
                'computational_cost': nb101_result.computational_cost,
            }
        except Exception as e:
            logger.error(f"NAS-Bench-101 evaluation failed: {e}")
            results['benchmarks']['nasbench101'] = {'error': str(e)}
        
        # NAS-Bench-201
        try:
            logger.info("Evaluating on NAS-Bench-201")
            nb201_result = self.benchmarks['nasbench201'].robustness_analysis(arch_dict)
            results['benchmarks']['nasbench201'] = nb201_result
        except Exception as e:
            logger.error(f"NAS-Bench-201 evaluation failed: {e}")
            results['benchmarks']['nasbench201'] = {'error': str(e)}
        
        # NAS-Bench-301
        try:
            logger.info("Evaluating on NAS-Bench-301")
            nb301_result = self.benchmarks['nasbench301'].predict_performance(arch_dict)
            results['benchmarks']['nasbench301'] = nb301_result
        except Exception as e:
            logger.error(f"NAS-Bench-301 evaluation failed: {e}")
            results['benchmarks']['nasbench301'] = {'error': str(e)}
        
        # HPOBench
        try:
            logger.info("Evaluating on HPOBench")
            hpo_result = self.benchmarks['hpobench'].evaluate_anytime(arch_dict)
            results['benchmarks']['hpobench'] = hpo_result
        except Exception as e:
            logger.error(f"HPOBench evaluation failed: {e}")
            results['benchmarks']['hpobench'] = {'error': str(e)}
        
        # NATS-Bench
        try:
            logger.info("Evaluating on NATS-Bench")
            nats_scaling = self.benchmarks['nats_bench'].evaluate_scaling(arch_dict)
            nats_transfer = self.benchmarks['nats_bench'].evaluate_transfer(arch_dict)
            results['benchmarks']['nats_bench'] = {
                'scaling': nats_scaling,
                'transfer': nats_transfer,
            }
        except Exception as e:
            logger.error(f"NATS-Bench evaluation failed: {e}")
            results['benchmarks']['nats_bench'] = {'error': str(e)}
        
        self.evaluation_history.append(results)
        logger.info("Comprehensive evaluation complete")
        
        return results
    
    def evaluate_population(self, architectures: List[Dict]) -> Dict:
        """
        Evaluate multiple architectures
        
        Args:
            architectures: List of DPO-NAS architecture dictionaries
            
        Returns:
            Dictionary with individual results and summary statistics
        """
        logger.info(f"Evaluating population of {len(architectures)} architectures")
        
        individual_results = []
        for idx, arch in enumerate(architectures):
            logger.info(f"Evaluating architecture {idx + 1}/{len(architectures)}")
            result = self.evaluate_architecture(arch)
            individual_results.append(result)
        
        summary = self._compute_summary(individual_results)
        
        return {
            'individual_results': individual_results,
            'summary': summary,
            'population_size': len(architectures),
        }
    
    def _compute_summary(self, evaluations: List[Dict]) -> Dict:
        """
        Compute summary statistics across population
        
        Args:
            evaluations: List of evaluation results
            
        Returns:
            Dictionary with aggregated statistics
        """
        summary = {}
        
        # NAS-Bench-101 summary
        nb101_accs = []
        for eval_result in evaluations:
            nb101 = eval_result['benchmarks'].get('nasbench101', {})
            if 'test_accuracy' in nb101:
                nb101_accs.append(nb101['test_accuracy'])
        
        if nb101_accs:
            summary['nasbench101'] = {
                'mean_accuracy': float(np.mean(nb101_accs)),
                'std_accuracy': float(np.std(nb101_accs)),
                'max_accuracy': float(np.max(nb101_accs)),
                'min_accuracy': float(np.min(nb101_accs)),
            }
        
        # NAS-Bench-201 summary
        nb201_accs = []
        for eval_result in evaluations:
            nb201 = eval_result['benchmarks'].get('nasbench201', {})
            if 'mean_accuracy' in nb201:
                nb201_accs.append(nb201['mean_accuracy'])
        
        if nb201_accs:
            summary['nasbench201'] = {
                'mean_accuracy': float(np.mean(nb201_accs)),
                'std_accuracy': float(np.std(nb201_accs)),
                'max_accuracy': float(np.max(nb201_accs)),
                'min_accuracy': float(np.min(nb201_accs)),
            }
        
        # NAS-Bench-301 summary
        nb301_accs = []
        for eval_result in evaluations:
            nb301 = eval_result['benchmarks'].get('nasbench301', {})
            if 'predicted_accuracy' in nb301:
                nb301_accs.append(nb301['predicted_accuracy'])
        
        if nb301_accs:
            summary['nasbench301'] = {
                'mean_accuracy': float(np.mean(nb301_accs)),
                'std_accuracy': float(np.std(nb301_accs)),
                'max_accuracy': float(np.max(nb301_accs)),
                'min_accuracy': float(np.min(nb301_accs)),
            }
        
        # HPOBench summary
        hpo_accs = []
        for eval_result in evaluations:
            hpo = eval_result['benchmarks'].get('hpobench', {})
            if 'final_accuracy' in hpo:
                hpo_accs.append(hpo['final_accuracy'])
        
        if hpo_accs:
            summary['hpobench'] = {
                'mean_accuracy': float(np.mean(hpo_accs)),
                'std_accuracy': float(np.std(hpo_accs)),
                'max_accuracy': float(np.max(hpo_accs)),
                'min_accuracy': float(np.min(hpo_accs)),
            }
        
        # NATS-Bench summary
        nats_accs = []
        for eval_result in evaluations:
            nats = eval_result['benchmarks'].get('nats_bench', {})
            if 'scaling' in nats and 'scaling_analysis' in nats['scaling']:
                nats_accs.append(nats['scaling']['scaling_analysis']['mean_accuracy'])
        
        if nats_accs:
            summary['nats_bench'] = {
                'mean_accuracy': float(np.mean(nats_accs)),
                'std_accuracy': float(np.std(nats_accs)),
                'max_accuracy': float(np.max(nats_accs)),
                'min_accuracy': float(np.min(nats_accs)),
            }
        
        return summary
    
    def save_results(self, filepath: str) -> None:
        """
        Save evaluation results to JSON file
        
        Args:
            filepath: Path to output JSON file
        """
        output_path = Path(filepath)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        with open(output_path, 'w') as f:
            json.dump(self.evaluation_history, f, indent=2, default=str)
        
        logger.info(f"Results saved to {filepath}")
    
    def generate_report(self, output_dir: str) -> str:
        """
        Generate Markdown report of evaluations
        
        Args:
            output_dir: Directory to save the report
            
        Returns:
            Path to generated report
        """
        output_path = Path(output_dir)
        output_path.mkdir(parents=True, exist_ok=True)
        
        report_file = output_path / f"benchmark_report_{datetime.now().strftime('%Y%m%d_%H%M%S')}.md"
        
        with open(report_file, 'w') as f:
            f.write("# DPO-NAS Comprehensive Benchmark Report\n\n")
            f.write(f"Generated: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n\n")
            f.write(f"Total Evaluations: {len(self.evaluation_history)}\n\n")
            
            if self.evaluation_history:
                f.write("## Summary Statistics\n\n")
                summary = self._compute_summary(self.evaluation_history)
                
                f.write("| Benchmark | Mean Accuracy | Std | Min | Max |\n")
                f.write("|-----------|---------------|-----|-----|-----|\n")
                
                for bench_name, stats in summary.items():
                    if 'mean_accuracy' in stats:
                        f.write(f"| {bench_name} | {stats['mean_accuracy']:.4f} | "
                               f"{stats['std_accuracy']:.4f} | {stats['min_accuracy']:.4f} | "
                               f"{stats['max_accuracy']:.4f} |\n")
                
                f.write("\n## Individual Results\n\n")
                for idx, result in enumerate(self.evaluation_history):
                    f.write(f"### Architecture {idx + 1}\n\n")
                    f.write(f"Timestamp: {result['timestamp']}\n\n")
                    
                    for bench_name, bench_result in result['benchmarks'].items():
                        f.write(f"**{bench_name}**\n")
                        if 'error' not in bench_result:
                            f.write(f"```json\n{json.dumps(bench_result, indent=2)}\n```\n\n")
                        else:
                            f.write(f"Error: {bench_result['error']}\n\n")
        
        logger.info(f"Report generated: {report_file}")
        return str(report_file)



from typing import Dict, Tuple
from dpo import DPO_NAS
from dpo.core.config import DPO_Config
from dpo.evaluation.ensemble import EnsembleEstimator

class MyCustomEstimator:
    def estimate(self, arch_dict: Dict) -> Tuple[float, Dict]:
        loss = 0.3
        metrics = {
            'latency_ms': 30.0,
            'memory_mb': 20.0,
            'flops_m': 100.0,
        }
        return loss, metrics

if __name__ == "__main__":
    config = DPO_Config.fast()
    ensemble = EnsembleEstimator([])
    ensemble.estimators['custom'] = MyCustomEstimator()
    def estimate(arch_dict):
        return ensemble.estimators['custom'].estimate(arch_dict)
    # Monkey patch simple single-estimator behavior
    ensemble.estimate = estimate  # type: ignore
    optimizer = DPO_NAS(config, estimator=ensemble)
    results = optimizer.optimize()
    print("Best Fitness:", results['best_fitness'])



from typing import Dict, Optional, Tuple

class EvaluationCache:
    def __init__(self, max_size: int = 10000):
        self.cache: Dict[str, Tuple[float, Dict]] = {}
        self.access_count: Dict[str, int] = {}
        self.max_size = max_size

    def get(self, arch_hash: str) -> Optional[Tuple[float, Dict]]:
        if arch_hash in self.cache:
            self.access_count[arch_hash] += 1
            return self.cache[arch_hash]
        return None

    def put(self, arch_hash: str, result: Tuple[float, Dict]) -> None:
        if len(self.cache) >= self.max_size:
            lru_key = min(self.access_count, key=self.access_count.get)
            del self.cache[lru_key]
            del self.access_count[lru_key]
        self.cache[arch_hash] = result
        self.access_count[arch_hash] = 1

    def clear(self) -> None:
        self.cache.clear()
        self.access_count.clear()





import numpy as np
import logging
from typing import Dict, Optional
from collections import deque

from .config import DPO_Config
from .agent import SearchAgent
from ..architecture.gene import ArchitectureGene
from ..evaluation.ensemble import EnsembleEstimator
from ..constraints.handler import AdvancedConstraintHandler
from ..utils.logger import get_logger

class DPO_NAS:
    def __init__(self,
                 config: DPO_Config,
                 estimator: Optional[EnsembleEstimator] = None,
                 constraint_handler: Optional[AdvancedConstraintHandler] = None,
                 logger: Optional[logging.Logger] = None):
        self.config = config
        self.estimator = estimator or EnsembleEstimator()
        self.constraint_handler = constraint_handler or AdvancedConstraintHandler(config)
        self.logger = logger or get_logger('DPO-NAS', self.config.verbose)
        self.islands = [[] for _ in range(config.num_islands if config.island_model else 1)]
        self.best_agent: Optional[SearchAgent] = None
        self.best_fitness: float = float('inf')
        self.best_per_island = [None] * len(self.islands)
        self.history = {
            'iterations': [],
            'best_fitness': [],
            'avg_fitness': [],
            'worst_fitness': [],
            'improvement_rate': [],
            'best_architecture': None,
        }
        self.improvement_tracker = deque(maxlen=30)
        
        # Pre-calculate bounds for enforcement
        self._ops_len = len(ArchitectureGene.OPERATIONS) - 0.01
        self._kers_len = len(ArchitectureGene.KERNELS) - 0.01

    def initialize_population(self) -> None:
        agents_per_island = max(1, self.config.population_size // len(self.islands))
        self.logger.info(f"Initializing population: {self.config.population_size} agents")
        
        for island_idx, island in enumerate(self.islands):
            for _ in range(agents_per_island):
                gene = ArchitectureGene()
                # Create agent
                agent = SearchAgent(gene=gene, island_id=island_idx)
                # Evaluate
                self._evaluate_agent(agent, iteration=0)
                island.append(agent)
                
                # Update bests
                if agent.fitness < self.best_fitness:
                    self.best_fitness = agent.fitness
                    self.best_agent = agent
                
                if self.best_per_island[island_idx] is None or agent.fitness < self.best_per_island[island_idx].fitness:
                    self.best_per_island[island_idx] = agent

    def _evaluate_agent(self, agent: SearchAgent, iteration: int) -> None:
        arch_dict = agent.gene.to_architecture_dict()
        loss, metrics = self.estimator.estimate(arch_dict)
        penalty = self.constraint_handler.compute_adaptive_penalty(metrics, iteration, self.config.max_iterations)
        
        # Fast arithmetic
        lat_norm = metrics['latency_ms'] / self.config.latency_constraint
        mem_norm = metrics['memory_mb'] / self.config.memory_constraint
        flop_norm = metrics['flops_m'] / self.config.flops_constraint
        
        agent.fitness = (
            self.config.w_loss * loss +
            self.config.w_latency * lat_norm +
            self.config.w_memory * mem_norm +
            self.config.w_flops * flop_norm +
            penalty
        )
        agent.metrics = metrics

    def _get_adaptive_parameters(self, iteration: int) -> Dict[str, float]:
        t = iteration / max(1, self.config.max_iterations)
        
        # Vectorized-style scalar operations
        decay = (1 - t) ** self.config.decay_power
        alpha = self.config.alpha_0 * decay
        beta = self.config.beta_0 + 0.5 * t
        gamma = self.config.gamma_0 * (1 - 0.5 * t)
        delta = self.config.delta_0 * (1 - 0.3 * t)
        
        if self.config.adaptive_alpha and len(self.improvement_tracker) > 10:
            # sum / len is faster than np.mean for deque
            it = self.improvement_tracker
            recent_avg = sum(list(it)[-10:]) / 10.0
            alpha *= 1.2 if recent_avg < 0.001 else 0.9
            
        return {
            'alpha': float(np.clip(alpha, 0.01, 0.3)),
            'beta': float(np.clip(beta, 0.8, 1.5)),
            'gamma': float(np.clip(gamma, 0.5, 1.5)),
            'delta': float(np.clip(delta, 0.05, 0.4)),
        }

    def _clip_gene_array(self, gene_arr: np.ndarray, num_layers: int, num_cells: int) -> None:
        """Helper to enforce bounds in-place on numpy array"""
        # Operations
        np.clip(gene_arr[:num_layers], 0, self._ops_len, out=gene_arr[:num_layers])
        # Kernels
        np.clip(gene_arr[num_layers:2*num_layers], 0, self._kers_len, out=gene_arr[num_layers:2*num_layers])
        # Skip connections
        np.clip(gene_arr[2*num_layers:2*num_layers+num_cells], 0, 1, out=gene_arr[2*num_layers:2*num_layers+num_cells])
        # Multipliers
        np.clip(gene_arr[-2:], 0.3, 2.0, out=gene_arr[-2:])

    def _dpo_step(self, agent: SearchAgent, params: Dict, iteration: int) -> SearchAgent:
        # OPTIMIZED: Perform vector math on arrays without creating intermediate ArchitectureGene objects
        # This reduces object overhead significantly
        
        current_vec = agent.gene.gene
        
        # 1. Debt Gene Logic (Agent + Perturbation)
        perturbation = np.random.randn(agent.gene.D) * params['alpha']
        debt_vec = current_vec + perturbation
        self._clip_gene_array(debt_vec, agent.gene.num_layers, agent.gene.num_cells)
        
        # Debt Vector
        debt_delta = debt_vec - current_vec
        
        # 2. Repay Gene Logic (Agent - Beta * Debt)
        repay_vec = current_vec - (params['beta'] * debt_delta)
        self._clip_gene_array(repay_vec, agent.gene.num_layers, agent.gene.num_cells)
        
        # 3. Double Gene Logic (Repay - Gamma * Debt)
        double_vec = repay_vec - (params['gamma'] * debt_delta)
        self._clip_gene_array(double_vec, agent.gene.num_layers, agent.gene.num_cells)
        
        # 4. Smart Gene Logic (Double + Delta * (GlobalBest - Current))
        best_global = self.best_agent or agent
        smart_vec = double_vec + params['delta'] * (best_global.gene.gene - current_vec)
        self._clip_gene_array(smart_vec, agent.gene.num_layers, agent.gene.num_cells)
        
        # Create Final Object
        smart_gene = ArchitectureGene(agent.gene.num_layers, agent.gene.num_cells)
        smart_gene.gene = smart_vec
        
        candidate = SearchAgent(gene=smart_gene, island_id=agent.island_id)
        self._evaluate_agent(candidate, iteration)
        
        if candidate.fitness < agent.fitness:
            candidate.improvements = agent.improvements + 1
            return candidate
            
        agent.age += 1
        return agent

    def _migrate_between_islands(self) -> None:
        if not self.config.island_model or len(self.islands) < 2:
            return
            
        num_migrate = max(1, self.config.population_size // 20)
        
        for island_idx in range(len(self.islands)):
            # Find best neighbors
            best_neighbors = [
                self.best_per_island[o] 
                for o in range(len(self.islands)) 
                if o != island_idx and self.best_per_island[o]
            ]
            
            if best_neighbors:
                best_neighbor = min(best_neighbors, key=lambda a: a.fitness)
                # Sort current island
                self.islands[island_idx].sort(key=lambda a: a.fitness, reverse=True) # Worst first
                
                # Replace worst
                for i in range(num_migrate):
                    if i < len(self.islands[island_idx]):
                        # Deep copy gene
                        new_gene = best_neighbor.gene.copy()
                        new_agent = SearchAgent(gene=new_gene, island_id=island_idx)
                        self._evaluate_agent(new_agent, iteration=0)
                        self.islands[island_idx][i] = new_agent

    def _inject_diversity(self) -> None:
        for island in self.islands:
            n = len(island)
            elite_count = max(1, int(n * self.config.elite_ratio))
            island.sort(key=lambda a: a.fitness)
            elite = island[:elite_count]
            
            # Replace non-elites
            for i in range(elite_count, n):
                if elite:
                    parent = elite[np.random.randint(0, len(elite))]
                    mutant_gene = parent.gene.mutate()
                    new_agent = SearchAgent(gene=mutant_gene, island_id=parent.island_id)
                    self._evaluate_agent(new_agent, iteration=0)
                    island[i] = new_agent

    def optimize(self) -> Dict:
        self.logger.info("=" * 70)
        self.logger.info("Starting DPO-NAS Optimization")
        self.logger.info(f"Strategy: {self.config.eval_strategy}")
        self.logger.info(f"Population: {self.config.population_size} | Islands: {len(self.islands)}")
        self.logger.info("=" * 70)
        
        self.initialize_population()
        
        for iteration in range(self.config.max_iterations):
            params = self._get_adaptive_parameters(iteration)
            
            for island in self.islands:
                for idx, agent in enumerate(island):
                    new_agent = self._dpo_step(agent, params, iteration)
                    island[idx] = new_agent
                    
                    if new_agent.fitness < self.best_fitness:
                        self.best_fitness = new_agent.fitness
                        self.best_agent = new_agent
                        self.logger.info(f"[Iter {iteration}] New best: {self.best_fitness:.4f}")

            # Periodic tasks
            if (iteration + 1) % self.config.migration_freq == 0 and self.config.island_model:
                self._migrate_between_islands()
                
            if (iteration + 1) % self.config.diversity_inject_freq == 0:
                self._inject_diversity()

            # Tracking
            # Optim: generator expression for fitness sum/stats
            all_fitness = np.array([a.fitness for island in self.islands for a in island])
            
            improvement = (self.history['best_fitness'][-1] - self.best_fitness) if self.history['best_fitness'] else 0.0
            self.improvement_tracker.append(improvement)
            
            self.history['iterations'].append(iteration)
            self.history['best_fitness'].append(self.best_fitness)
            self.history['avg_fitness'].append(float(np.mean(all_fitness)))
            self.history['worst_fitness'].append(float(np.max(all_fitness)))
            self.history['improvement_rate'].append(float(improvement))
            
            if iteration % 10 == 0:
                self.logger.info(f"[Iter {iteration:3d}] Best: {self.best_fitness:.4f} | Avg: {np.mean(all_fitness):.4f} | Improvement: {improvement:.6f}")

        if self.best_agent:
            self.history['best_architecture'] = self.best_agent.gene.to_architecture_dict()

        self.logger.info("=" * 70)
        self.logger.info("Optimization Complete!")
        self.logger.info(f"Best Fitness: {self.best_fitness:.4f}")
        self.logger.info("=" * 70)

        return {
            'best_fitness': self.best_fitness,
            'best_architecture': self.history.get('best_architecture'),
            'best_metrics': getattr(self.best_agent, 'metrics', {}),
            'history': self.history,
            'config': {
                'population_size': self.config.population_size,
                'islands': len(self.islands),
                'eval_strategy': self.config.eval_strategy,
                'max_iterations': len(self.history['iterations']),
            }
        }